/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: ldb.c,v 1.66 2003/03/24 14:34:35 kripke Exp $
	implementation of general db access functions.
*/

#include <stdlib.h>
#include <string.h>
#include <limits.h> /* PATH_MAX */
#include <errno.h>

/* gettimeofday */
#ifdef WIN32
#   include <sys/timeb.h>
#   include <sys/types.h>
#   define timeval _timeb
#else
#   include <unistd.h>
#endif
#include <sys/time.h>


/* special */
#if defined( __GNUC__ ) && defined ( alloca )
#include <alloca.h>
#endif

/* gcc defines always a cpu type - this we use for byteorder checking */
#if defined( sparc ) || defined( __ppc__ )
#	define LDB_BIG_ENDIAN
/* TODO: figure out fastest "htonl" on those boxes that usually don't swap */
static int rvi ( int i ) {
	int r;
	((char*)&r)[0] = ((char*)&i)[3];
	((char*)&r)[1] = ((char*)&i)[2];
	((char*)&r)[2] = ((char*)&i)[1];
	((char*)&r)[3] = ((char*)&i)[0];
	return r;
}
static short rvs ( short i ) {
	short r;
	((char*)&r)[0] = ((char*)&i)[1];
	((char*)&r)[1] = ((char*)&i)[0];
	return r;
}
#define SWI( i ) i = rvi( i )
#define SWS( s ) s = rvs( s )
#else
# define rvi(i) i
# define rvs(s) s
#define SWI( i )
#define SWS( s )
#endif
#if defined( sparc )
#	define LDB_NEEDALIGN
#endif

#include "lstr.h"
#include "lio.h"
#include "lbt.h"
#include "lcs.h"
#include "ldb.h"
#include "lfdt.h"
#include "luti.h"



/* ************************************************************
	private types
*/

/** extension of master file proper. */
static const char EXT_MST_MST[] = ".mst";
/** extension of master file xref. */
static const char EXT_MST_XRF[] = ".xrf";

typedef enum {
	MST_MST,
	MST_XRF,
	MST_FILES
} mst_file;

static const char * const EXT_MST[MST_FILES] = {
	EXT_MST_MST,
	EXT_MST_XRF
};

/** extension of inverted file short term nodes. */
static const char EXT_INV_N01[] = ".n01";
/** extension of inverted file short term leaves. */
static const char EXT_INV_L01[] = ".l01";
/** extension of inverted file long term nodes. */
static const char EXT_INV_N02[] = ".n02";
/** extension of inverted file long term leaves. */
static const char EXT_INV_L02[] = ".l02";
/** extension of inverted file control records. */
static const char EXT_INV_CNT[] = ".cnt";
/** extension of inverted file postings. */
static const char EXT_INV_IFP[] = ".ifp";

typedef enum {
	INV_N01,
	INV_L01,
	INV_N02,
	INV_L02,
	INV_CNT,
	INV_IFP,
	INV_FILES
} inv_file;

static const char * const EXT_INV[INV_FILES] = {
	EXT_INV_N01,
	EXT_INV_L01,
	EXT_INV_N02,
	EXT_INV_L02,
	EXT_INV_CNT,
	EXT_INV_IFP
};

/** extension of lbt B-Link-Tree.
	It's named oxi because that is nicer than oix for OpenIsis indeX.
	however, see http://www.oxicenter.com.br/
*/
static const char EXT_LBT_OXI[] = ".oxi";
static const char * const EXT_LBT[] = {
	EXT_LBT_OXI
};

/** extension of supporting file alpha character table. */
static const char EXT_SUP_ACT[] = ".act";
/** extension of supporting file uppercase table. */
static const char EXT_SUP_UCT[] = ".uct";

typedef enum {
	SUP_ACT,
	SUP_UCT,
	SUP_FILES
} sup_file;

static const char * const EXT_SUP[SUP_FILES] = {
	EXT_SUP_ACT,
	EXT_SUP_UCT
};


typedef int lblk[128];


typedef struct {
	Db          head;
	int         flags;
	const char *path;
	int         mst[MST_FILES]; /* master file */
	int         inv[INV_FILES]; /* primary inverted file */
	int         mfc[LMFC__FL]; /* master file control record */
	unsigned    mflen; /* master file length */
	int         xrf[129]; /* last used xrf block : THREAD THREAT */
	int         xrlen; /* length of xrf (in blocks) */
	lblk       *mmap; /* memory map of xrf */
	int         mmlen; /* length of map (in blocks) */
	int         cnt[LDB_INDEXES][LCNT__FL]; /* two cnt records */
	short       tlen[LDB_INDEXES]; /* max term length for each index */
	int         set[2]; /* record sets, should be { LSET_MST, LSET_INV } */
	LcsTab      ctab[LCS__TABS];
	Idx         oxi;
} LDb;


/* db flags */
#define DB_OPEN     0x010000
#define DB_INVOPEN  0x020000
#define DB_LBTOPEN  0x040000
#define DB_WRITABLE 0x080000
#define DB_MODIFIED 0x100000
#define DB_MMAP  0x10
#define DB_VARI 0xf /* mask for variant */

/* get master or inverted record set */
#ifdef HAVE_NOSTANDARD_RSETS /* far future */
#	define DB_RSET( db, s ) lstrlib[ (db)->set[s] ]
#else /* simply LSET_MST and LSET_INV */
#	define DB_RSET( db, s ) lstrlib[ s ]
#endif

/* get xstr for record rec in set */
#define DB_XSTR( db, set, rec ) \
	DB_RSET( db, set ).desc[ DB_VARI & (db)->flags ][ rec ]
/* get record names for record rec in set */
#define DB_RNAM( db, set, rec ) \
	DB_RSET( db, set ).name[ rec ]


/** packed little endian masterfile control structure.
*/
typedef struct Mfc {
	int   ctlm;
	int   nmfn;
	int   nmfb;
	short nmfp;
	short type;
	int   rcnt;
	int   mfx1;
	int   mfx2;
	int   mfx3;
} Mfc;


/** packed little endian masterfile record.
*/
typedef struct Dict {
	short tag;
	short pos;
	short len;
} Dict;

/** packed little endian masterfile record.
*/
typedef struct Mfr {
	int   mfn;
	short recl; /* a.k.a. mfrl */
	short bwbl; /* low part of int */
	short bwbh; /* high part of int */
	short bwp;
	/* it is believed, that this first five fields up to here (12 bytes packed)
	are to be in one 512-byte block; the manual mentiones even 14 bytes ... ???
	*/
	short base;
	short nvf;
	short stat;
	Dict  dict[1];
} Mfr;



/* ************************************************************
	private data
*/

static LDb defdbspace[32];
/* array of open dbs. should expand dynamically. */
static LDb *dbs = defdbspace;
static int dbs_len = sizeof(defdbspace)/sizeof(defdbspace[0]);

static int init;


/* ************************************************************
	private functions
*/
static LDb *getDb ( int id )
{
	if ( 0 <= id && id < dbs_len && dbs[id].flags ) {
		return &dbs[id];
	}
	log_msg( LOG_ERROR, "attempt to access bad db id %d", id );
	return 0;
}	/* getDb */


/* ************************************************************
	start of io section
*/
enum {
	OPEN_WR,
	OPEN_RD,
	OPEN_OPT
};

/* set extension. fname MUST already end with .xxx.
*/
static char *setext ( char *fname, const char *ext )
{
	int l = strlen( fname ) - 4;
	memcpy( fname+l, ext, 4 );
	if ( 'A'<=fname[l-1] && fname[l-1]<= 'Z' ) {
		char *p = fname+l;
		for ( ;*p; p++ ) /* use uppercase extensions */
			if ( 'a' <= *p && *p <= 'z' )
				*p -= 'a'-'A';
	}
	return fname;
}

static int openfiles ( int *fid, char *path,
	const char * const *ext, int nfiles, int readonly )
{
	int mode = LIO_SEEK | (readonly ? LIO_RD : LIO_RDWR|LIO_CREAT|LIO_FLOCK);
	int i,ret;

	for ( i=0; i<nfiles; i++ ) {
		setext( path, ext[i] );
		fid[i] = lio_open( path, mode );
		log_msg( LOG_INFO, "opening file '%s' %c 0x%x",
			path, readonly ? 'r' : 'w', fid[i] );
		if ( 0 <= fid[i] )
			continue;
		fid[i] = 0;
		if ( OPEN_OPT <= readonly )
			continue;
		ret = log_msg( LOG_SYSERR, "could not open file '%s' for %sing",
			path, readonly ? "read" : "writ" );
		while ( i-- )
			lio_close( &fid[i], LIO_INOUT );
		return ret; /* bad */
	}
	return 0; /* good */
}	/* openfiles */


static int closefiles ( int *fid, int nfiles )
{
	int ret = 0, i;
	for ( i=0; i<nfiles; i++ )
		lio_close( &fid[i], LIO_INOUT );
	return ret;
}	/* closefiles */


static int readblk ( void *dst, int siz, int fid, int where )
{
	int got;
	got = lio_pread( &fid, dst, siz, where );
	if ( 0 > got )
		return got;
#ifndef NDEBUG
	if ( LOG_DO( LOG_ALL ) )
		LOG_HEX( dst, got );
#endif
	if ( siz == (int)got )
		return 0;
	log_msg( LOG_WARN, "got %u bytes wanted %d at %d in 0x%x",
		got, siz, where, fid );
	return 1+(int)got;
}	/* readblk */


/* ************************************************************
	end of io section
*/

static int *nrec ( int *xstr )
{
	int *dst = (int*)lmalloc( xstr[LSTR_ILEN] );
	if ( dst )
		*dst = *xstr;
	return dst;
}	/* nrec */



typedef struct b8 { char x[8]; } b8;
typedef struct b4 { char x[4]; } b4;
typedef struct b2 { char x[2]; } b2;


static int convert ( int *dst, char *src, int *xstr )
{
	static int pow2[] = { 1, 2, 4, 8 };
	int occ = -1;
	int *xmbrs = xstr+LSTR_XMBR;
	int nmbrs  = LSTRFIX(*xstr);
	int *mbr;
	char *buf, *part = src, *srcend = src + xstr[LSTR_XLEN];

	if ( ! dst )
		return log_msg( ERR_NOMEM, "out of memory (no dst) in convert" );
	if ( LSTRLEN(*xstr) > xstr[LSTR_ILEN] )
		return log_msg( ERR_INVAL, "bad ilen %d need %d",
			xstr[LSTR_ILEN], (int)LSTRLEN(*xstr) );
	if ( *dst != *xstr )
		return log_msg( ERR_INVAL, "bad *dst 0x%08x need 0x%08x", *dst, *xstr );
	/* clean and re-init */
	memset( dst, 0, xstr[LSTR_ILEN] );
	*dst = *xstr;
	mbr  = dst+1;
	buf  = ((char*)dst) + LSTRLEN(*dst);

	/* cvt the fixed part (occ==-1) and each occurrence of repeated part. */
	for ( ;/* occ < LSTROCC(*dst) */; ) { /* cvt one part */
		int i;
		/* TODO: this loop is crucial for performance */
		for ( i=0; i<nmbrs; i++, mbr++ ) { /* assign one xmbr */
			int xmbr = xmbrs[i];
			char *s = part + LONG2OFF(xmbr);
			int sbytes = srcend - s;
			int len,j;
			union {
				char buf[8];
				b8   x8;
				b4   x4;
				b2   x2;
				short s;
				int  i;
				lll  ll;
			} num;
			if ( LMBRISNUM( xmbr ) ) {
				/* numeric data */
#if defined(LDB_BIG_ENDIAN )
#	define NEEDSWAP(xmbr) ( ! ( LMBR_FHIE & (xmbr) ) )
#else
#	define NEEDSWAP(xmbr) (LMBR_FHIE & (xmbr))
#endif
				int ld = LMBRLD(xmbr);
#ifndef LDB_NEEDALIGN
				if ( ! NEEDSWAP( xmbr ) ) {
					/* much faster this way ... */
					/* TODO: len and bit checks; actually not needed yet ... */
					switch ( ld ) {
					case 3: *mbr = *(lll*)s; break;
					case 2: *mbr = *(int*)s; break;
					case 1: *mbr = *(short*)s; break;
					case 0: *mbr = *s; break;
					}
					continue;
				}
#endif
				len = pow2[ LMBRLD(xmbr) ];
				if ( len > sbytes )
					return log_msg( ERR_INVAL,
						"srcbuf too short %d have %d need %d occ %d mbr %d",
						xstr[LSTR_XLEN], sbytes, len, occ, i );
				if ( !NEEDSWAP( xmbr ) )
					/* for ( j = len; j--; ) num.buf[j] = s[j]; */
					switch ( ld ) {
					case 3: num.x8 = *(b8*)s; break;
					case 2: num.x4 = *(b4*)s; break;
					case 1: num.x2 = *(b2*)s; break;
					case 0: num.buf[0] = *s; break;
					}
				else /* swap bytes */
					for ( j = len; j--; )
						num.buf[j] = s[len - 1 - j];
				switch ( len ) {
				case 8: *mbr = num.ll; break;
				/* TODO: defines for 16 and 64 bit compilers */
				case 4: *mbr = num.i; break;
				case 2: *mbr = num.s; break;
				case 1: *mbr = num.buf[0]; break;
				}
				if ( LMBRISBITS( xmbr ) ) { /* apply bit shift and mask */
					*mbr >>= LMBRBITOFF( xmbr );
					*mbr &= ~(-1L << LMBRBITLEN( xmbr ));
				}
				continue;
			} 
			/* else raw data -- that's easy :) */
			{
				int offset = buf - (char*)dst;
				int need;
				len = LONG2LEN(xmbr);
				need = offset + len;
				if ( need > xstr[LSTR_ILEN] )
					return log_msg( ERR_INVAL,
						"bad buflen %d need %d+%d in occ %d mbr %d",
						xstr[LSTR_ILEN], offset, len, occ, i );
				if ( len > sbytes )
					return log_msg( ERR_INVAL,
						"srcbuf too short %d have %d need %d occ %d mbr %d",
						xstr[LSTR_XLEN], sbytes, len, occ, i );
				memcpy( buf, s, len );
				*mbr = buf - (char*)dst;
				buf += len;
			}
		}	/* for mbrs */

		if ( ++occ >= LSTROCC(*dst) )
			break;
		if ( occ )
			part += (short)xstr[LSTR_XRLO]; /* adv. rep. part len */
		else { /* was the fixed part, setup for repeated */
			nmbrs = LSTRREP(*xstr);
			part  += (short)(xstr[LSTR_XRLO]>>16); /* adv. rep. part off */
			xmbrs += i;
		}
	}
	return 0;
}	/* convert */


static int readrec ( int *dst, int fid, int where, int *xstr )
{
	char *buf = (char *)
#ifdef alloca
		alloca( xstr[LSTR_XLEN] )
#else
		lmalloc( xstr[LSTR_XLEN] )
#endif
	;
	int got = 0;
	int ret = 0;

	if ( ! buf )
		return -ERR_NOMEM;
	if ( 0 > where )
		where = -where * xstr[LSTR_XLEN];
	got = readblk( buf, xstr[LSTR_XLEN], fid, where );
	ret = got ? got : convert( dst, buf, xstr );
#ifndef alloca
	free( buf );
#endif

	return ret;
}	/* readrec */


/* read and log */
static int readlog ( int *dst, int fid, int where, LDb *db, int set, int rec )
{
	int ret = readrec( dst, fid, where, DB_XSTR( db, set, rec ) );
	if ( !ret && LOG_DO( LOG_TRACE ) )
		LOG_STR( dst, lstrlib[ db->set[set] ].name[ rec ] );
	return ret;
}	/* readlog */



/* ************************************************************
	access functions for the record types
*/

static int getOff ( LDb *db, int rowid, int xr )
{
	int rowix = rowid - 1; /* mfns count from 1 */
	int xrf_block = rowix / 127;
	int blkix = xrf_block + 1; /* ... so do xrf blocks */
	int off;
	if ( xrf_block < db->mmlen ) {
		if ( xr )
			db->mmap[xrf_block][1+(int)(rowix % 127)] = rvi( xr );
		else
			xr = rvi( db->mmap[xrf_block][1+(int)(rowix % 127)] );
	} else {
		int err = 0;
		/* if ( LIO_LOCK() ) return -3; */
		if ( xr ) { /* write */
			if ( blkix <= db->xrlen ) {
				SWI( xr );
				if ( 4 != lio_pwrite( &db->mst[MST_XRF], &xr, 4,
					xrf_block*512 + 4*(1 + (rowix % 127))) )
					return 0;
			} else { /* extent */
				lblk extend;
				memset( extend, 0, sizeof(extend) );
				while ( db->xrlen < blkix ) { /* extend */
					extend[0] = db->xrlen + 1; /* set blk id */
					if ( blkix == extend[0] )
						extend[1+(int)(rowix % 127)] = rvi( xr );
					SWI( extend[0] );
					if ( 512 != lio_pwrite( &db->mst[MST_XRF], extend, 512,
						db->xrlen*512) )
						return 0;
					db->xrlen++;
				}
			}
			/* and go on read it back, just to check ... */
		}
		if ( blkix != db->xrf[LXRF_XPOS] ) {
			int ret;
			LOG_DBG( LOG_VERBOSE, "fetching xrf block %d had %d",
				blkix, ! db->xrf ? -1 : db->xrf[LXRF_XPOS] );
			ret = readlog( db->xrf, db->mst[MST_XRF],
				-xrf_block, db, LSET_MST, LSTR_XRF );
			if ( ret ) {
				log_msg( LOG_ERROR, "\twhen fetching xrf block %d", blkix );
				err = -1;
			} else if ( blkix == -db->xrf[LXRF_XPOS] ) {
				LOG_DBG( LOG_DEBUG, "hmmm ... negative" );
				db->xrf[LXRF_XPOS] = blkix;
			} else if ( blkix != db->xrf[LXRF_XPOS] ) {
				log_msg( LOG_WARN, "bad xrf %d wanted %d",
					db->xrf[LXRF_XPOS], blkix );
				err = -2;
			}
		}
		xr = db->xrf[ LXRF_XREC + (int)(rowix % 127) ];
		/* LIO_RELE(); */
		if ( err )
			return err;
	}
	/*
	21 bits (<<11) signed for the (512 byte) block ("xrmfb")
		1 for the first block (offset 0)
		0 means, never had such a record
		-1 and xrmfp=0: record removed from MST
		(there is no record at pos 0 in 1st block,
		since there resides the MST header)
		other negative value -x or pos!=0:
		record logically deleted, was at +x
	1 bit (1<<10): this record is new and not yet inverted
	1 bit (1<<9): this record is changed and not yet re-inverted
	9 bits for the block-relative position ("xrmfp")
	*/
	off = (((xr & 0xfffff800) >> 2) - 0x200) | (0x1ff & xr);
	if ( 0 < (xr & ~0x600) ) {
		LOG_DBG( LOG_DEBUG,
			"offset for rowid %d is %d (blk %d pos %d) flg 0x%08x at %d[%d]",
			rowid, off, (xr>>11)&0xfffff, xr&0x1ff, xr&0x80000600, blkix, rowix%127 );
		return off;
	}
	/* deleted */
	log_msg( LOG_INFO,
		"offset for rowid %d is %d (blk %d pos %d) flg 0x%08x at %d[%d]",
		rowid, off, (xr>>11)&0xfffff, xr&0x1ff, xr&0x80000600, blkix, rowix%127 );
	return 0;
}	/* getOff */


static int* getMfr ( LDb *db, int off, int *nxtoff )
{
	struct mfcxstr {
		int xstr[LSTR_LONGS(7+3)];
	} my = *(struct mfcxstr *)DB_XSTR( db, LSET_MST, LSTR_MFR ); 
	int head[1+7+3];
	int len = 0, base;
	int *rec = 0;
	char *buf = 0;
#ifdef alloca
	int notalloca = 0;
#endif

	LOG_DBG( LOG_VERBOSE, "getting MFR at off %d", off );
	if ( 498 < off % 512 )
		log_msg( LOG_WARN, "blk pos > 498 in offset 0x%08x", off );
	*head = *my.xstr;
	if ( readrec( head, db->mst[MST_MST], off, my.xstr ) ) {
		log_msg( LOG_ERROR, "\twhen reading MFR head at %d", off );
		return 0;
	}
	/* log_str( LOG_VERBOSE, head, lstrlib[LSET_MST].name[LSTR_MFR] ); */
	len = head[LMFR_RECL];
	LOG_DBG( LOG_VERBOSE, "got MFR %d reclen %d", head[LMFR_MFN], len );
	if ( nxtoff ) {
		*nxtoff = off + (0 < len ? len : -len); /* an odditiy */
		if ( 1 & *nxtoff ) /* an odditiy */
			(*nxtoff)++; /* round up to even */
		if ( 498 < *nxtoff % 512 ) {
			*nxtoff += 512;
			*nxtoff &= ~0x1ff;
		}
		if ( *nxtoff > db->mfc[LMFC_NMFB]*512 + db->mfc[LMFC_NMFP] ) {
			LOG_DBG( LOG_VERBOSE, "at end of db: %d > %d*512+%hd",
				*nxtoff, db->mfc[LMFC_NMFB], db->mfc[LMFC_NMFP] );
			*nxtoff = -1;
		}
	}
	if ( len < 0 ) {
		log_msg( LOG_INFO, "found deleted rec len %hd at offset %d", len, off );
		return 0;
	}

	/* check external base length */
	base = LONG2OFF(my.xstr[LSTR_XRLO])
		+ head[LMFR_NVF]*LONG2LEN(my.xstr[LSTR_XRLO]);
	if ( 0 > head[LMFR_BASE] || 0 > head[LMFR_NVF]
		|| 0x8fff < head[LMFR_NVF]
		|| len < head[LMFR_BASE] || len < base
		|| (head[LMFR_NVF] && head[LMFR_BASE] < base)
	) {
		log_msg( LOG_ERROR,
			"bad len %d base %d nvf %d need base %d at offset %d",
			len, head[LMFR_BASE], head[LMFR_NVF], base, off );
		/* check alignment problem */
		base = LONG2OFF(my.xstr[LSTR_XRLO])
			+ head[LMFR_STAT]*LONG2LEN(my.xstr[LSTR_XRLO]);
		if ( 0 > head[LMFR_NVF] || 0 > head[LMFR_STAT]
			|| 0x8fff < head[LMFR_STAT]
			|| len < head[LMFR_NVF] || len < base
			|| (head[LMFR_STAT] && head[LMFR_NVF] < base)
		) 
			;
		else
			log_msg( LOG_ERROR, "probably alignment problem, try -format aligned" );
		goto cleanup;
	}

	/* set up external structure for this rec */
	my.xstr[LSTR_SIZE] |= head[LMFR_NVF] << 16; /* occ of rep. part */
	my.xstr[LSTR_XLEN] = len;
	/* internal base length */
	base = LSTRLEN( *my.xstr );
	/* internal len adjusted for the slightly longer base */
	my.xstr[LSTR_ILEN] = len + base - head[LMFR_BASE]; /* the buffer */

	rec = nrec( my.xstr );
	if ( ! rec ) {
		log_msg( LOG_SYSERR, "could not alloc MFR of len %hd", my.xstr[LSTR_ILEN] );
		goto cleanup;
	}
	buf = (char *)
#ifdef alloca
		alloca( len );
	/* first try faster alloca, but stack may be too limited for large records */
	notalloca = ! buf;
	if ( notalloca )
		buf =
#endif
		lmalloc( len );

	if ( ! buf ) {
		log_msg( LOG_SYSERR, "could not alloc MFR of len %hd", len );
		goto cleanup;
	}
	if ( readblk( buf, len, db->mst[MST_MST], off ) ) {
		log_msg( LOG_ERROR, "\twhen reading MFR" );
		goto cleanup;
	}
#ifndef LDB_BIG_ENDIAN
	if ( LVAR_PAC != (DB_VARI & db->flags) ) {
#endif
		if ( convert( rec, buf, my.xstr ) ) {
			log_msg( LOG_ERROR, "\twhen converting MFR" );
			goto cleanup;
		}
#ifndef LDB_BIG_ENDIAN
	} else { /* 10% faster */
		Mfr *mfr = (Mfr*)buf;
		short *s = &mfr->dict->tag;
		int *f = rec + LMFR__FL;
		int *e = f + 3*head[LMFR_NVF];
		rec[LMFR_MFN]  = mfr->mfn;
		rec[LMFR_RECL] = mfr->recl;
		/*
		rec[LMFR_BWB]  = mfr->bwbh<<16 | mfr->bwbl;
		rec[LMFR_BWP]  = mfr->bwp;
		*/
		rec[LMFR_BASE] = mfr->base;
		rec[LMFR_NVF]  = mfr->nvf;
		rec[LMFR_STAT] = mfr->stat;
		while ( f < e ) {
			*f++ = *s++;
			*f++ = *s++;
			*f++ = *s++;
		}
	}
#endif

	if ( rec[LMFR_STAT] ) {
		log_msg( LOG_WARN, "found status %hd", rec[LMFR_STAT] );
		goto cleanok;
	}

	/* do a consistency check */
	if ( rec[LMFR_NVF] < 0 || rec[LMFR_BASE] < 0 ) {
		log_msg( LOG_ERROR, "found neg. field nvf %hd base %hd",
			rec[LMFR_NVF], rec[LMFR_BASE] );
		goto cleanup;
	}

	/* now care for the field values */
	{
		char *valsrc = buf+rec[LMFR_BASE];
		char *recsta = ((char*)rec);
		char *valdst = recsta + base;
		int xbufl   = rec[LMFR_RECL] - rec[LMFR_BASE];
		int sumlens = 0;
		int i;
		for ( i=0; i < rec[LMFR_NVF]; i++ ) {
			int *d = &rec[LMFR__FL + i*LMFR__RL];
			if ( d[LMFR_POS] < 0 || d[LMFR_LEN] < 0 ) {
				log_msg( LOG_ERROR,
					"bad field %d at off %d: negativ pos %hd or len %hd",
					i, off,  d[LMFR_POS], d[LMFR_LEN] );
				goto cleanup;
			}
			if ( d[LMFR_POS] + d[LMFR_LEN] > xbufl ) {
				log_msg( LOG_ERROR,
					"bad field %d at off %d: pos %hd + len %hd > buf %d",
					i, off,  d[LMFR_POS], d[LMFR_LEN], xbufl );
				goto cleanup;
			}
			sumlens += d[LMFR_LEN];
			if ( sumlens > xbufl ) {
				log_msg( LOG_ERROR,
					"bad fields at off %d: sum of lengths %d > buf %d",
					off,  sumlens, xbufl );
				goto cleanup;
			}
			memcpy( valdst, valsrc+d[LMFR_POS], d[LMFR_LEN] );
			d[LMFR_POS] = valdst - recsta;
			valdst += d[LMFR_LEN];
		}
	}	/* consistency check */
	rec[LMFR_BWB] = /* "used" bytes */
	rec[LMFR_RECL] = my.xstr[LSTR_ILEN];
	rec[LMFR_BWP] = rec[LMFR_NVF]; /* avail fields = used fields */
	rec[LMFR_BASE] = base;

	if ( LOG_TRACE <= log_lev )
		LOG_STR( rec, lstrlib[LSET_MST].name[LSTR_MFR] );
	goto done;

cleanup:
	if ( nxtoff )
		*nxtoff = -1;
cleanok:
	if ( rec ) {
		free( rec );
		rec = 0;
	}
done:
	if ( buf
#ifdef alloca
		&& notalloca
#endif
	)
		free( buf );
	if ( rec )
		*rec = db->head.dbid;
	return rec;
}	/* getMfr */


/** write the record.
	If it doesn't yet have a mfn, assign one.
	NOTE: on a BIG_ENDIAN, anything but the mfn and recl will be frobbed
	after this call
*/
static int putMfr ( LDb *db, Mfr *mfr )
{
	int oldpos, newpos;
	int ret;

	if ( !(db->flags & DB_WRITABLE) ) {
		log_msg( LOG_ERROR, "db is not writable" );
		return -1;
	}
	/* minimalist sanity check */
	if ( mfr->nvf < 0
		|| mfr->base != 18 + 6*mfr->nvf
		|| mfr->base > mfr->recl
	) {
		log_msg( LOG_ERROR, "bad nvf/base/recl %d/%d/%d ",
			mfr->nvf, mfr->base, mfr->recl );
		return -2;
	}
	db->flags |= DB_MODIFIED;
	if ( mfr->mfn ) {
		int block;
		if ( db->mfc[LMFC_NMFN] > mfr->mfn )
			oldpos = getOff( db, mfr->mfn, 0 );
		else {
			db->mfc[LMFC_NMFN] = mfr->mfn+1;
			oldpos = 0;
		}
		mfr->bwp = 511 & oldpos;
		block = 1 + (oldpos >> 9); /* blockno counting from 1 */
		mfr->bwbl = 0xffff & block;
		mfr->bwbh = block >> 16;
		if ( db->mfc[LMFC_NMFN] <= mfr->mfn )
			db->mfc[LMFC_NMFN] = mfr->mfn+1;
	} else {
		mfr->mfn = db->mfc[LMFC_NMFN]++; /* assign new mfn */
		oldpos = 0;
		mfr->bwbl = mfr->bwbh = mfr->bwp = 0;
	}
	mfr->stat = 0;
	newpos = db->mflen;
	if ( 498 < (newpos & 511) ) /* round up to next block boundary */
		newpos = ~511 & (newpos + 14);
	if ( 1 & newpos )
		newpos++;
	db->mflen = newpos + mfr->recl;
#ifdef LDB_BIG_ENDIAN
{ /* swap swap swap */
	Dict *d = mfr->dict;
	short nvf = mfr->nvf;
	SWI( mfr->mfn ); SWS( mfr->recl ); SWS( mfr->bwbl ); SWS( mfr->bwbh );
	SWS( mfr->bwp ); SWS( mfr->base ); SWS( mfr->nvf  ); SWS( mfr->stat );
	for ( ; nvf--; d++ ) {
		SWS( d->tag );
		SWS( d->pos );
		SWS( d->len );
	}
}
#endif
	ret = lio_pwrite( &db->mst[MST_MST], (char*)mfr, rvs(mfr->recl), newpos );
#ifdef LDB_BIG_ENDIAN
	/* restore mnf, recl */
	SWI( mfr->mfn );
	SWS( mfr->recl );
#endif
	if ( ret != mfr->recl )
		return log_msg( ERR_TRASH, "could not write Mfr %d bytes got %d",
			mfr->recl, ret );
	getOff( db, mfr->mfn, (1 << (oldpos ? 9 : 10))
		| (((newpos & 0xfffffe00) + 0x200) << 2) | (0x1ff & newpos) );

	return 0;
}	/* putMfr */


static int putRec ( LDb *db, Rec *rec )
{
	int ret = 0, i;
	Mfr *mfr = 0;
	int buflen = 0;
	int reclen = 0;
	int contig = 1;
#ifdef alloca
	int notalloca = 0;
#endif
	const char *rbase = ((char *)rec) + rec->base;
	Dict *d;
	Field *f = rec->field;
	/* TODO: if not rec->len, delete ? */
	for ( i = rec->len; i--; f++ ) {
		if ( ! f->len )
			continue;
		if ( ! f->val )
			LOG_OTO( cleanup, ( ERR_FAULT, "bad rec NULL val" ) );
		contig = contig && (f->val == rbase + buflen);
		buflen += f->len;
	}
	reclen = 18 + 6*rec->len + buflen;
	if ( 1 & reclen )
		reclen++;
	mfr = (Mfr*)
#ifdef alloca
		alloca( reclen );
	notalloca = ! mfr;
	if ( notalloca )
		mfr = (Mfr*)
#endif
		lmalloc( reclen );
	if ( ! mfr )
		LOG_OTO( cleanup,
			( ERR_NOMEM, "could not alloc MFR of len %hd", reclen ) );
	mfr->mfn  = rec->rowid;
	mfr->recl = reclen;
	mfr->bwbl = mfr->bwbh = mfr->bwp = 0;
	mfr->base = 18 + 6*rec->len;
	mfr->nvf  = rec->len;
	mfr->stat = 0;
	d = mfr->dict;
	f = rec->field;
	buflen = 0;
	for ( i = rec->len; i--; d++, f++ ) {
		d->tag = f->tag;
		d->pos = buflen;
		buflen += (d->len = f->len);
	}
	if ( contig )
		memcpy( ((char*)mfr)+mfr->base, rbase, buflen );
	else {
		char *mbase = ((char*)mfr)+mfr->base;
		d = mfr->dict;
		f = rec->field;
		for ( i = rec->len; i--; d++, f++ )
			if ( d->len )
				memcpy( mbase + d->pos, f->val, d->len );
	}
	ret = putMfr( db, mfr );
	if ( !ret && !rec->rowid )
		rec->rowid = mfr->mfn;

cleanup:
	if ( mfr
#ifdef alloca
		&& notalloca
#endif
	)
		free( mfr );
	return ret;
}	/* putRec */



/* ************************************************************
	utilities
*/

static int search ( LDb *db, const char *key, LdbPost *post,
	Rec *rec, DXLoop *lp )
{
	int i, j, prefix, idx, ret, ock;
	int pos;
	int *leaf, *entry;
	char *term;
	int *xstr;
	struct { /* terms cursor */
		char  key[LDB_MAX_KEYLEN+1]; /* key or key prefix */
		short klen; /* key length to compare */
		char  imin; /* minimum index to search */
		char  imax; /* maximum index to search */
		int   leaf[LDB_INDEXES][LDB_TERMBUF]; /* one leaf buffer per index */
		short lpos[LDB_INDEXES]; /* next position in leaf, -1 if done */
	} crs;
	short klen; /* length for initial locate */
	int block[128]; /* buffer to read one block */
	int blockpos = 0;
	LdbP *p = 0;

	if ( ! key )
		key = "$";
	/* prepare cursor struct */
	memset( &crs, 0, sizeof(crs) ); /* tabula rasa */
	crs.klen = strlen( key );
	/* check for prefix match */
	if ( post )
		prefix = LDB_PFX & post->mode;
	else if ( (prefix = crs.klen && '$' == key[crs.klen - 1]) )
		crs.klen--;
	/* check out minimum index to search */
	for ( crs.imin=0; crs.klen > db->tlen[(int)crs.imin]; )
		if ( LDB_INDEXES == ++(crs.imin) )
			return log_msg( ERR_INVAL, "bad keylen %d key '%.64s'", crs.klen, key );
	/* prepare key */
	memset( crs.key, ' ', sizeof(crs.key)-1 );
	{
		unsigned char *uc = (unsigned char*)crs.key;
		unsigned char *uk = (unsigned char*)key;
		for ( i=crs.klen; i--; )
			uc[i] = db->ctab[LCS_UCASE].c[ uk[i] ];
	}
	if ( prefix )
		crs.imax = LDB_INDEXES-1;
	else {
		crs.imax = crs.imin;
		crs.klen = db->tlen[(int)crs.imin];
	}
	log_msg( LOG_INFO, "search for '%.*s'%c", crs.klen, crs.key, prefix?'$':' ' );
	key = crs.key;
	klen = crs.klen;
	if ( rec && rec->len ) {
		/* use last key from record to locate starting position */
		key = rec->field[rec->len-1].val;
		klen = rec->field[rec->len-1].len;
		rec->len = 0;
	}

	for ( i=crs.imin; i<=crs.imax; i++ ) { /* find leaf positions */
		int nFile = INV_N01 + 2*i; /* node file index */
		int nStr = LSTR_N01 + 2*i; /* node struct index */
		int *nstr = DB_XSTR( db, LSET_INV, nStr );
		int lvl;
		short cmplen = klen <= db->tlen[i] ? klen : db->tlen[i];
		pos = db->cnt[i][LCNT_POSR]; /* pos of root record */
		j = 0;
		for ( lvl = 0; 0<pos; lvl++ ) { /* traverse node levels */
			int node[102];
			LOG_DBG( LOG_DEBUG, "node %d at %d lvl %d", pos, j, lvl );
			assert( (int)sizeof(node) >= nstr[LSTR_ILEN] );
			*node = *nstr;
			ret = readlog( node, db->inv[nFile], 1-pos, db, LSET_INV, nStr );
			if ( pos != node[LN0X_POS] /* wrong address */
				|| i+1 != node[LN0X_TYPE] /* wrong type */
				|| node[LN0X_OCK] < 1 /* no keys */
				|| 2*db->cnt[i][LCNT_ORDN] < node[LN0X_OCK] /* too many keys */
			)
				return log_msg( ERR_TRASH, "bad node pos %d type %d keys %d",
					node[LN0X_POS], node[LN0X_TYPE], node[LN0X_OCK]
				);
			ock = node[LN0X_OCK];
			for ( j=1;
				j<ock && 0 < (ret = memcmp( key,
					((char*)node)+node[j*LN0X__RL+LN0X__FL+LN0X_KEY], cmplen ));
				j++ )
				;
			/* now j is at end or on next index not less */
			if ( j==ock /* end */
				|| ret /* index is greater than key */
				|| prefix /* backtrack even on exact match */
			)
				j--; /* step into last ock with lower key */
			pos = node[LN0X__FL + j*LN0X__RL + LN0X_REF];
		} /* for lvl */
		/* got some negative ref to leaf; set leaf pos */
		crs.leaf[i][LL0X_PS] = -pos;
		/*
		since the lpos and LL0X_OCK are both 0 by the memset above,
		we will initially load the leaves
		*/
	} /* for indexes */
	/* done preparing cursor */

	if ( post ) /* prepare for postings */
		p = post->p;
	xstr = DB_XSTR( db, LSET_INV, LSTR_IFP ); 

	for (;;) { /* loop terms in prefix mode */
		/* vars for postings: */
		int infb, infp; /* block and pos where to read postings */
		int added; /* postings added or marked per term */
		int blkno; /* postings block number */
		int remain = 0; /* postings to fetch from next block of segment */
		int ifp[LIFP__FL]; /* postings header */

		idx = -1; /* index to use */
		/* compare index terms, load leafes if needed */
		for ( i = crs.imin; i <= crs.imax; i++ ) {
			short cmplen = klen <= db->tlen[i] ? klen : db->tlen[i];
			leaf = crs.leaf[i];
			if ( leaf[LL0X_OCK] <= crs.lpos[i] ) { /* load */
				int lFile = INV_L01 + 2*i; /* leaf file index */
				int lStr = LSTR_L01 + 2*i; /* leaf struct index */
				int *lstr = DB_XSTR( db, LSET_INV, lStr );

				crs.lpos[i] = -1;
			reread:
				if ( ! (pos = leaf[LL0X_PS]) )
					continue;
				LOG_DBG( LOG_DEBUG, "leaf %d", pos );
				assert( (int)sizeof(crs.leaf[i]) >= lstr[LSTR_ILEN] );
				*leaf = *lstr;
				ret = readlog( leaf, db->inv[lFile], 1-pos, db, LSET_INV, lStr );
				if ( pos != leaf[LL0X_POS] /* wrong address */
					|| i+1 != leaf[LL0X_TYPE] /* wrong type */
					|| leaf[LL0X_OCK] < 1 /* no keys */
					|| 2*db->cnt[i][LCNT_ORDN] < leaf[LL0X_OCK] /* too many keys */
				)
					return log_msg( ERR_TRASH, "bad leaf pos %d type %d keys %d",
						leaf[LL0X_POS], leaf[LL0X_TYPE], leaf[LL0X_OCK] );
				ock = leaf[LL0X_OCK];
				/* advance to first term which is not too small
					(should be needed only for first leaf of an index)
				*/
				for ( j=0;
					j<ock && (0 < (ret = memcmp( key,
						((char*)leaf)+leaf[LL0X__FL + j*LL0X__RL + LL0X_KEY], cmplen ))
						|| (!ret && key!=crs.key) ); /* skip exact while locating */
					j++ )
					;
				if ( ock == j )
					goto reread; /* start over w/ next leaf of same index */
				if ( 0 <= ret
					|| (key!=crs.key && !memcmp( crs.key,
						((char*)leaf)+leaf[LL0X__FL + j*LL0X__RL + LL0X_KEY], crs.klen ))
				)
					crs.lpos[i] = j;
				/* else let -1 */
			} /* if reload */
			if ( 0 > crs.lpos[i] )
				continue;
			if ( 0 > idx ) {
				idx = i;
				continue;
			}
			/* compare this index next term to that of index idx */
			/* assume that index w/ lower number has shorter keys */
			ret = memcmp(
				((char*)leaf)+leaf[LL0X__FL + crs.lpos[i]*LL0X__RL + LL0X_KEY],
				((char*)crs.leaf[idx])+
				crs.leaf[idx][LL0X__FL + crs.lpos[idx]*LL0X__RL + LL0X_KEY],
				db->tlen[idx] );
			if ( 0 > ret )
				idx = i;
		}
		if ( 0 > idx )
			goto done;
		j = crs.lpos[idx];
		leaf = crs.leaf[idx];
		entry = leaf + LL0X__FL + j*LL0X__RL;
		term = ((char*)leaf) + entry[LL0X_KEY];
		if ( memcmp( crs.key, term, crs.klen ) )
			goto done;
		crs.lpos[idx]++;

		if ( rec ) { /* record the term */
			/* field to assign */
			Field *f = rec->field + rec->len;
			short tlen = db->tlen[idx];
			/* end of available buffer */
			char *b = rec->len
				? (char*)f[-1].val /* before previously assigned field */
				: ((char*)rec + rec->bytes); /* end of record */
			while ( tlen && ' ' == term[tlen-1] )
				tlen--;
			b -= tlen;
			if ( b < (char*)(f+1) ) /* no space left on device */
				goto done;
			/* probably we're nuking the locator now: */
			memcpy( b, term, tlen );
			f->tag = 0;
			f->val = b;
			f->len = tlen;
			rec->len++;
			/* reset key from locator to prefix */
			key = crs.key;
			klen = crs.klen;
		}

		if ( ! post && ! lp )
			continue;
		/* collect postings */
		infb = entry[LL0X_INFB];
		infp = entry[LL0X_INFP];
		/* the IFP file is organized in blocks of 128 longs.
			1st int is block number followed by 127 data.
			postings are organized in chained segments so that each segment
			fits within one such block. a segment has five longs header,
			giving number of postings and pointer to next segment.
		*/
		added = 0;
		for ( blkno=0; infb; blkno++ ) { /* segments */
			LdbP merge[127/2]; /* buffer to collect new postings */
			int *base; /* start of data */
			int *b; /* start of postings */
			int n; /* max postings in this seg's 1st block */
			int xlen; /* external length to read */
			int f = post ? post->fil - 1 : 0; /* highest pos to consider in given postings */
			int m = 0; /* fill merge buffer */
			int k; /* loop segment */

			if ( infp > 127-2-5 ) {
				return log_msg( ERR_TRASH, "found bad IFP pos %d blk %d for %.*s",
					infp, blkno, klen, key );
			}
			if ( remain ) { /* consecutive block of same segment */
				n = remain;
				if ( n > 127/2 )
					n = 127/2;
				xlen = 8*n;
			} else {
				n = (127 - 5 - infp)/2;
				xlen = 20 + 8*n;
			}
			pos = (infb - 1) * 512 + (infp + 1) * 4;
			if ( blockpos
				&& !((pos-blockpos) >> 9) /* 0 <= (pos-blockpos) < 512 */
				&& pos+xlen <= blockpos+ 1 + (0x1ff & ~blockpos)
			)
				base = block + (pos - blockpos)/sizeof(int);
			else {
				int blklen = 1 + (0x1ff & ~pos);
				assert( xlen <= blklen );
				assert( blklen <= (int)sizeof(block) );
				assert( 0 == (0x1ff & (pos + blklen)) );
				ret = readblk( block, blklen, db->inv[INV_IFP], pos );
				if ( ret )
					return log_msg( ERR_IO, "\twhen reading IFP" );
				blockpos = pos;
				base = block;
			}
			if ( remain ) { /* no header to convert */
				remain -= n;
				b = base; /* no header */
			} else  {
				assert( (int)sizeof(ifp) >= xstr[LSTR_ILEN] );
				*ifp = *xstr;
				ret = convert( ifp, (char *)base, xstr );
				if ( ret )
					return log_msg( ERR_TRASH, "\twhen converting IFP header" );
				if ( n > ifp[LIFP_SEGP] )
					n = ifp[LIFP_SEGP];
				remain = ifp[LIFP_SEGP] - n;
				b = base+5; /* after header */
			}
			LOG_DBG( LOG_VERBOSE,
				"key %d.%d '%.*s' blk %d post %d/%d r %d xlen %d at b/p %d.%d=%d",
				leaf[LL0X_PS], j, db->tlen[idx], term, blkno,
				n, ifp[LIFP_TOTP], remain, xlen, infb, infp, pos );
			if ( LOG_DO( LOG_TRACE ) )
				LOG_STR( ifp, lstrlib[ db->set[LSET_INV] ].name[ LSTR_IFP ] );
			assert( (size_t)n <= sizeof(merge)/sizeof(merge[0]) );
			if ( lp ) {
				Key kbf;
				Hit hit;
				unsigned char tlen = (unsigned char) db->tlen[idx];
				while ( tlen && ' ' == term[tlen-1] )
					tlen--;
				memcpy( kbf.byt, term, kbf.len = tlen );
				for ( k=0; k<n; k++ ) { /* callback needs 'em sorted */
					int ppos;
					unsigned char *c = (unsigned char *)&b[k*2];
					LdbP e; /* the entry */
#if defined(LDB_BIG_ENDIAN )
					memcpy(e.bytes,c,8);
#else
					e.bytes[0] = c[7];	e.bytes[1] = c[6];
					e.bytes[2] = c[5];	e.bytes[3] = c[4];
					e.bytes[4] = c[3];	e.bytes[5] = c[2];
					e.bytes[6] = c[1];	e.bytes[7] = c[0];
#endif
					ppos = LDBP_POS( &e );
					hit.mfn = (unsigned)LDBP_ROW( &e );
					hit.tag = (unsigned short)LDBP_TAG( &e );
					hit.occ = (unsigned short)(ppos >> 16);
					hit.pos = (unsigned short)ppos;
					if ( lp->cb( lp->me, &kbf, &hit ) )
						goto done;
				}
			}
			if ( post ) for ( k=n; k--; ) {
				/* loop backwards (for the fun of it) postings in segment */
				int prow, ptag, ppos;
				unsigned char *c = (unsigned char *)&b[k*2];
				LdbP e; /* the entry */
				LdbP samerow; /* highest possible entry w/ same row as e */
#if defined(LDB_BIG_ENDIAN )
				/* the 8 bytes of a posting are BIG ENDIAN ! */
				memcpy(e.bytes,c,8);
#else
				e.bytes[0] = c[7];	e.bytes[1] = c[6];
				e.bytes[2] = c[5];	e.bytes[3] = c[4];
				e.bytes[4] = c[3];	e.bytes[5] = c[2];
				e.bytes[6] = c[1];	e.bytes[7] = c[0];
#endif
				prow = LDBP_ROW( &e );
				ptag = LDBP_TAG( &e );
				ppos = LDBP_POS( &e );
				LOG_DBG( LOG_VERBOSE, "post %d.%hd pos %06x key '%.*s'",
					prow, ptag, ppos, db->tlen[idx], term );
				if ( 0 >= ptag /* bad tag */
					|| !prow || prow >= db->mfc[LMFC_NMFN] /* bad mfn */
				)
					continue;
				if ( ! post
					|| (post->cut && prow >= post->cut)
					|| (post->tag && post->tag != ptag)
				)
					continue;
				if ( prow < post->skp ) /* quickly bail out on skip condition */
					break;
				LDBP_SETROWTOP( &samerow, &e ); /* for mfn comparison */
				/* sweep down to postings for the same row as e ... */
				while ( f >= 0 && LDBP_GT( p+f, &samerow ) )
					f--;
				if ( LDB_AND & post->mode ) {
					int l;
					/* loop postings for same row, mark all (that are near enough) */
					LDBP_SETROWBOT( &samerow, &e ); /* for mfn comparison */
					/* NOTE: postings for row are GT than bottom even if marked */
					for ( l = f; l>=0 && LDBP_GT( p+l, &samerow ); l-- ) {
						if ( post->near ) {
							int dist;
							if ( ptag != LDBP_TAG( p+l ) ) continue;
							if ( LDB_NEAR_G != post->near ) {
								dist = LDBP_POS( p+l ) - LDBP_POS( &e );
								if ( dist < 0 ) dist = -dist;
								if ( 0 < post->near
									? post->near < dist
									: -post->near != dist /* exact $$$$ */
								) continue;
							}
						}
						LDBP_SETMARK( p+l );
						added++;
					}
				} else {	/* OR mode */
					int add;
					if ( ! post->near ) /* add if row not found: ignore details */
						add = 0 > f || prow > LDBP_ROW( p+f );
					else {	/* add if no exact match */
						int l;
						/* NOTE: we don't use mark bit in OR mode, do we ? */
						for ( l = f; l>=0 && LDBP_GT( p+l, &e ); l-- )
							;
						add = 0 > l || LDBP_GT( &e, p+l );
					}
					if ( add )
						merge[ m++ ] = e;
				}
			}	/* for postings in segment */
			if ( m ) { /* merge in the merge buffer */
				LdbP *mm = merge;
				added += m;
				for ( k = post->fil += m; m && k--; ) {
					LdbP src;
					if ( k < m || LDBP_GT( mm, &p[k-m] ) ) {
						src = *mm++;
						m--;
						LOG_DBG( LOG_DEBUG, "merging %d at %d", LDBP_ROW(&src), k );
					} else
						src = p[k-m];
					if ( k < post->len )
						p[k] = src;
					else { /* set cut */
						int row = LDBP_ROW( &src );
						if ( row < post->cut || !post->cut )
							post->cut = row;
					}
				}
				if ( post->fil > post->len )
					post->fil = post->len;
				if ( post->cut ) /* postings for cut row are unreliable */
					while ( post->fil && post->cut <= LDBP_ROW(p+post->fil-1) )
						post->fil--;
			}
			if ( remain ) { /* advance to start of next block */
				infb++;
				infp = 0;
			} else {
				infb = ifp[LIFP_NXTB];
				infp = ifp[LIFP_NXTP];
			}
		} /* for segments */
		LOG_DBG( LOG_VERBOSE, "added %d postings for key '%.*s'",
			added, db->tlen[idx], term );
	}	/* for terms in prefix/postings mode */
done:
	if ( post /* fixup */
		&& LDB_AND & post->mode && !(LDB_KEEPMARKS & post->mode)
	) {
		int mark = LDB_NOT & post->mode ? 0 : 0x8000;
		j=0;
		for ( i=0; i<post->fil; i++ )
			if ( mark == LDBP_MARK(p+i) ) {
				LDBP_CLRMARK(p+i);
				p[j++] = p[i];
			}
		post->fil = j;
	}
	return ! rec ? 0 : rec->len;
}	/* search */


static int is_abs_path (const char *path) {
	if (! path || ! *path) {
		return 0;
	}
	if ('/' == *path) {
		return 1;
	}
#ifdef WIN32
	if ('\\' == *path) {
		return 1;
	}
	if (':' == path[1] && ('\\' == path[2] || '/' == path[2])) {
		return 1;
	}
#endif
	return 0;
}


static int ldb_open (const char *dbname, Rec *dbpar, Rec *syspar)
{
	LDb ndb, *db;
	int plen, sz, dbid;
	int ret = 0, invret = 0, lbtret = 0, autoformat = 1, writable = 1;
	char buf[4096], *p = buf;
	char path[ PATH_MAX ];

	memset( &ndb, 0, sizeof(ndb) );
	/* these should be 0 by memsetting to 0 anyway ... */
	ndb.path = 0; ndb.mmap = 0;
	ndb.flags |= DB_MMAP; /* it mean's: we'll try */

	/* loglevel */ {
		int llev = rInt2(dbpar, syspar, OPENISIS_SLOGV, -1);
		if ( 0 <= llev )
			cLog( llev, 0 );
	}

	if ( dbname ) {
		if (DBNLEN <= strlen (dbname)) {
			strncpy (ndb.head.name, dbname, DBNLEN - 1) [DBNLEN - 1] = 0;
			log_msg( LOG_WARN, "truncating dbname '%s' to '%s'",
				dbname, ndb.head.name );
		} else
			strcpy (ndb.head.name, dbname);
	} else {
		if (! dbpar && ! syspar)
			return -ERR_FAULT;
		if (! rString2 ( dbpar, syspar, OPENISIS_DNAME, path, sizeof(path)))
			return -ERR_FAULT;
		if (DBNLEN <= strlen (path)) {
			strncpy (ndb.head.name, path, DBNLEN - 1) [DBNLEN - 1] = 0;
			log_msg( LOG_WARN, "truncating dbname '%s' to '%s'",
				path, ndb.head.name );
		} else
			strcpy (ndb.head.name, path);
	}

	/* more init AFTER honoring verbosity */
	if ( ! init ) {
		lstr_auto(0);
		init = !0;
	}

	for ( dbid=0; dbid<dbs_len; dbid++ ) {
		if ( dbs[dbid].flags &&
			!strcmp( ndb.head.name, dbs[dbid].head.name ) ) {
			log_msg( LOG_INFO, "reopening %d '%s'", dbid, ndb.head.name );
			return dbid;
		}
	}
	/* go for slot */
	if ( dbid == dbs_len )
		for ( dbid=0; dbid<dbs_len && dbs[dbid].flags; dbid++ )
			;
	if ( dbid == dbs_len )
		return -1;
	db = &dbs[dbid];
	/* got slot */
	*db = ndb;
	db->head.dbid = dbid;
	db->set[ LSET_MST ] = LSET_MST;
	db->set[ LSET_INV ] = LSET_INV;
	if (dbpar || syspar) {
		char fmtstr[32];
		if (rString2 (dbpar, syspar, OPENISIS_DTYPE, fmtstr, sizeof(fmtstr))) {
			if (! strcmp ("aligned", fmtstr)) {
				db->flags |= LVAR_ALI;
				autoformat = 0;
			}
			else if (! strcmp ("naligned", fmtstr)) {
				autoformat = 0;
			}
		}
	}

	/* preset record sizes */
	db->mfc[0] = *DB_XSTR( db, LSET_MST, LSTR_MFC );
	db->xrf[0] = *DB_XSTR( db, LSET_MST, LSTR_XRF );
	db->cnt[0][0] =
	db->cnt[1][0] = *DB_XSTR( db, LSET_INV, LSTR_CNT );
	/* isis-1 index term lengths */
	db->tlen[0] = 10;
	db->tlen[1] = 30;

	/* only the packed little endian ("DOS") format is writable
	test later ...
	if ( LVAR_PAC != (DB_VARI & db->flags) )
		writable = 0;
	*/

	/* prepare path */
	plen = strlen (db->head.name);
	*path = 0;
	if (dbpar || syspar) {
		char dbpath[PATH_MAX];
		int slen = 0, dlen = 0;
		if ( rString2 (dbpar, syspar, OPENISIS_DPATH, dbpath, sizeof(path)) ) {
			dlen = strlen (dbpath);
			if (! dlen || ! is_abs_path (dbpath)) {
				if (rString2 (dbpar, syspar, OPENISIS_SPATH,
					path, sizeof(path))) {
					slen = strlen (path);
				}
			}
		}
		if (sizeof(path) < (unsigned int)
			(slen + 1 + dlen + 1 + plen + 4 + 1)) {
			return log_msg( ERR_FAULT, "path(%d+%d) or filename(%d) too long",
				slen, dlen, plen);
		}
		if (slen) {
			path[slen++] = '/';
		}
		if (dlen) {
			memcpy (path + slen, dbpath, dlen);
			path[slen += dlen] = '/';
			++slen;
		}
		memcpy (path + slen, db->head.name, plen);
		path[slen += plen] = 0;
		plen = slen;
	}
	else {
		memcpy (path, db->head.name, plen);
		path[plen] = 0;
	}
	/* save params */
	db->path = strdup( path );
	memcpy( path+plen, ".???", 5 );

	if (dbpar || syspar) {
		int ro = rInt2 (dbpar, syspar, OPENISIS_DRO, -1);
		if (0 < ro) {
			writable = 0;
		}
	}

	/* open files */
	if ( writable
		&& !openfiles( db->mst, path, EXT_MST, MST_FILES, OPEN_WR )
	)
		ret = 0;
	else {
		writable = 0;
		ret = openfiles( db->mst, path, EXT_MST, MST_FILES, OPEN_RD );
		if ( ret )
			goto cleanup;
	}

	/* XXX: MMMOOO */
	if ( writable
		&& !openfiles( &db->oxi.fd, path, EXT_LBT, 1, OPEN_WR )
	)
		lbtret = 0;
	else {
		writable = 0;
		lbtret = openfiles( &db->oxi.fd, path, EXT_LBT, 1, OPEN_RD );
	}

	/* trad. index is never openend writable. */
	invret = openfiles( db->inv, path, EXT_INV, INV_FILES, OPEN_RD );

	if ( autoformat ) {
		if ( invret )
			log_msg( LOG_WARN, "cannot guess format -- no inverted file" );
		else {
			unsigned len = lio_size( db->inv[INV_CNT] );
			if ( 56L == len ) {
				db->flags |= LVAR_ALI;
				/* writable = 0; we do not write aligned format */
			} else if ( 52L != len )
				log_msg( LOG_WARN, "cannot guess format -- bad .cnt len %d", len );
			log_msg( LOG_INFO, "using autoformat %saligned for .cnt len %d",
				(db->flags & LVAR_ALI) ? "":"un", len );
		}
	}
	if ( (ret = readlog( db->mfc, db->mst[MST_MST], 0, db, LSET_MST, LSTR_MFC )) )
	{
		/* NEW goto cleanup; */
		memset( db->mfc, 0, sizeof(db->mfc) );
		db->mfc[LMFC_NMFN] = 1;
		db->mfc[LMFC_NMFB] = 1;
		db->mfc[LMFC_NMFP] = 64;
		db->mflen = 64;
	} else {
		int lastblock = (db->mflen = lio_size( db->mst[MST_MST] ))/512;
		if ( 511 & db->mflen ) lastblock++; /* counting from 1 */
		/*
		the next record's block should be either the last one we have
		or the next one to follow
		*/
		if ( db->mfc[LMFC_NMFB] != lastblock && db->mfc[LMFC_NMFB] != lastblock+1 )
			log_msg( LOG_VERBOSE, "NMFB mismatch: NMFB %d ~ %d",
				db->mfc[LMFC_NMFB], lastblock );
		/* set LOGICAL mf length */
		db->mflen = (db->mfc[LMFC_NMFB]-1)*512 + db->mfc[LMFC_NMFP];
	}
	db->xrlen = lio_size( db->mst[MST_XRF] ) / 512;
	if ( (DB_MMAP & db->flags)
		&& db->xrlen
		&& db->xrlen*512
		== lio_mmap( &db->mst[MST_XRF], (void**)&db->mmap, db->xrlen*512 )
	)
		db->mmlen = db->xrlen;
	else
		db->flags &= ~DB_MMAP;

	/* supporting files, ctables */
	if ( 0 >= (sz = lio_slurp( &p, sizeof(buf), setext(path,EXT_SUP_ACT), 1 ))
		|| lcs_mktab( db->ctab+LCS_CTYPE, p, sz, LCS_A )
	)
		memcpy( db->ctab+LCS_CTYPE, lcs_latin1_ct, sizeof(db->ctab[0]) );
	if ( 0 >= (sz = lio_slurp( &p, sizeof(buf), setext(path,EXT_SUP_UCT), 1 ))
		|| lcs_mktab( db->ctab+LCS_UCASE, p, sz, 0 )
	)
		memcpy( db->ctab+LCS_UCASE, lcs_latin1_uc, sizeof(db->ctab[0]) );
	/* fill header */
	{
		struct timeval tv;
		Fd  *fds;
		Rec *rec;
		int  num = fdtFromFile (path, &fds, &rec);
		if (0 < num) {
			db->head.fdt = cFdt (ses0, num, fds, rec);
		}
		num = gettimeofday (&tv, 0);
		if (num) {
			log_msg (LOG_WARN, "dbopen(%s): gettimeofday returns %d %d\n",
				db->head.name, num, errno);
			db->head.tms = 0;
		}
		else {
			db->head.tms = (int) tv.tv_sec;
		}
		if (dbpar) {
			db->head.cfg = rDup (ses0, dbpar, 0, 0);
		}
	}

	/* done */
	db->flags |= DB_OPEN;
	if ( writable )
		db->flags |= DB_WRITABLE;

	if (dbpar || syspar) {
		int dmp = rInt2 (dbpar, syspar, OPENISIS_DDUMP, -1);
		if (0 < dmp) {
			int off = 0;
			int *r;
			do {
				if ( (r = ldb_readRecAtOff(dbid,off,&off)) )
					free( r );
			} while ( 0 < off );
			exit(0);
		}
	}

	if ( writable )
		db->oxi.flg |= LBT_WRITE;
	{
		char *oxityp = getenv("OXITYP");
		int ot;
		if ( oxityp && 0 < (ot = atoi(oxityp)) && 4 > ot )
			db->oxi.typ = ot << 4;
	}
	if ( ! lbtret
		&& ! lbt_init( &db->oxi )
	)
		db->flags |= DB_LBTOPEN;
	if ( LVAR_PAC != (DB_VARI & db->flags) )
		writable = 0;

	if ( ! invret
		&& ! (ret = readlog( db->cnt[0], db->inv[INV_CNT],
			0, db, LSET_INV, LSTR_CNT ))
		&& ! (ret = readlog( db->cnt[1], db->inv[INV_CNT],
			-1, db, LSET_INV, LSTR_CNT ))
	)
		db->flags |= DB_INVOPEN;

	return dbid;

cleanup:
	/* cleanup ... */
	db->flags = 0; /* MW: creation mode? */
	return 0 > ret ? ret : ret ? -ret : -1;
}	/* ldb_open */


/* ************************************************************
	package data
*/



/* ************************************************************
	package functions
*/

int *ldb_readRecAtOff ( int dbid, lxref off, int *nxtoff )
{
	int *rec;
	LDb *db = getDb( dbid );
	if ( ! db ) {
		log_msg( LOG_ERROR, "\tat ldb_readRecAtOff" );
		return 0;
	}
	if ( 0 == off )
		off = 64;
	rec = getMfr( db, off, nxtoff );
	if ( ! rec )
		return 0;
	LOG_DBG( LOG_VERBOSE, "db %d off %d: got %hd bytes",
		dbid, off, !rec ? -1 : rec[LMFR_RECL] );
	return rec;
}	/* ldb_readRecAtOff */



int ldb_search ( int dbid, const char *key, LdbPost *post, Rec *rec )
{
	LDb *db = getDb( dbid );
	Key k;

	if ( ! db )
		return -ERR_BADF;
	if ( post ) { /* prepare for postings */
		if ( ! post->len )
			post->len = sizeof(post->p)/sizeof(post->p[0]); /* standard length */
		if ( LDB_NOT & post->mode )
			post->mode |= LDB_AND;
	}
	if ( DB_INVOPEN & db->flags )
		return search( db, key, post, rec, 0 );
	if ( !(DB_LBTOPEN & db->flags) )
		return -ERR_BADF;
	if ( db->oxi.bat )
		return -ERR_BUSY;
	memset( &k, 0, sizeof(k) );
	if ( ! key ) {
		k.byt[0] = '$';
		k.len = 1;
	} else {
		unsigned char *uk = (unsigned char*)key;
		int l = strlen( key );
		if ( l > 255 )
			l = 255;
		k.len = (unsigned char)l;
		while ( l-- )
			k.byt[l] = db->ctab[LCS_UCASE].c[ uk[l] ];
	}
	return lbt_search( &db->oxi, &k, post, rec );
} /* ldb_search */


int ldb_p2s ( Set *set, LdbPost *post )
{
	int *s = set->id;
	int last=0, max = set->len;
	int i;
	set->len = 0;
	if ( ! max )
		max = OPENISIS_SETLEN;
	max--;
	if ( !post->fil )
		return 0L;
	s[0] = LDBP_ROW(post->p);
	for ( i=1; i<post->fil && last < max; i++ ) {
		int row = LDBP_ROW(post->p+i);
		if ( s[last] != row )
			s[++last] = row;
	}
	return set->len = last+1;
}	/* ldb_p2s */


#if 0
LcsTab *ldb_tabs( int dbid )
{
	LDb *db = getDb( dbid );
	return ! db ? 0 : db->ctab;
}	/* ldb_tabs */
#endif


Db *ldb_getdb (int dbid) {
	LDb *db = getDb (dbid);
	return db ? &db->head : 0;
}

/* ************************************************************
	public functions
*/
int dMaxId ( Session ses, int dbid )
{
	LDb *db = getDb( dbid );
	(void)ses;
	if ( ! db )
		return -ERR_BADF;
	return db->mfc[LMFC_NMFN] - 1;
}	/* dMaxId */


Raw *dRaw ( Session ses, int dbid, int rowid )
{
	int off;
	int *rec = 0;
	LDb *db;
	(void)ses;
	if ( LIO_LOCK() ) return 0;
	db = getDb( dbid );
	if ( ! db ) {
		log_msg( LOG_ERROR, "\tat openIsisReadRaw %d", rowid );
		goto done;
	}
	off = getOff( db, rowid, 0 );
	log_msg( LOG_INFO, "found xref 0x%08x for %d", off, rowid );
	if ( 0 >= off ) {
		log_msg( LOG_INFO, "found deleted xref 0x%08x for %d", off, rowid );
		goto done;
	}
	rec = getMfr( db, off, 0 );
	if ( ! rec ) {
		log_msg( LOG_WARN, "\tno record at %d rowid %d", off, rowid );
		goto done;
	}
	LOG_DBG( LOG_VERBOSE, "db %d row %d: got %hd bytes",
		dbid, rowid, !rec ? -1 : rec[LMFR_RECL] );
	if ( rec[LMFR_MFN] != rowid ) {
		log_msg( LOG_ERROR, "got mfn %d expected %d", rec[LMFR_MFN], rowid );
		free( rec );
		goto done;
	}
done:
	(void)LIO_RELE();
	return (Raw*)rec;
}	/* dRaw */


int dWritex ( Session ses, int dbid, Rec *rec, Rec *idx )
{
	LDb *db = getDb( dbid );
	int ret = 0;
	(void)ses;
	if ( ! db )
		return -ERR_BADF;
	if ( !(DB_WRITABLE & db->flags) )
		return log_msg( ERR_INVAL, "db %d not writable", dbid );
	if ( rec && (ret = putRec( db, rec )) )
		return ret;
	if ( idx ) {
		const unsigned char *const uc = db->ctab[LCS_UCASE].c;
		Hit h;
		Key k;
		Field *f = idx->field, *last = f + idx->len - 1;
		for ( ; f <= last; f++ ) {
			int tag = f->tag;
			int del = 0 > tag;
			if ( del )
				tag = -tag;
			else if ( ! tag )
				continue; /* ??? */
			memset( &k, 0, sizeof(k) ); /* we're so paranoid ! */
			switch ( tag ) {
			case XVAL: /* binary val */
				/* if ( f->len < db->oxi.vsz )
					memset( k.val.byt, 0, db->oxi.vsz - f->len );
				*/
				memcpy( k.val.byt
					+ (f->len < (int)db->oxi.vsz ? (int)db->oxi.vsz - f->len : 0),
					f->val, f->len > (int)db->oxi.vsz ? (int)db->oxi.vsz : f->len );
				k.val.len = db->oxi.vsz;
				break;
			case XHIT:
				if ( sizeof(Hit) != f->len )
					return log_msg( ERR_INVAL, "bad length %d for XHIT need %d",
						f->len, sizeof(Hit) );
				memcpy( &h, f->val, sizeof(Hit) );
				if ( !h.mfn && rec )
					h.mfn = rec->rowid;
				break;
			case XDOT: {
				int i = 0, v[5], *pv = v;
				const char *p = f->val, *e = p + f->len;
				for ( ; p < e && i<5; i++ ) {
					int dig = a2il( p, e-p, v+i );
					p += dig;
					if ( ! dig || (e > p && '.' != *p) )
						return log_msg( ERR_INVAL, "bad XDOT '%.*s'", e-p, p );
					p++;
				}
				memset( &h, 0, sizeof(h) );
				switch ( i ) {
				case 5: h.dbn = (unsigned short)*pv++;
				case 4: h.mfn = (unsigned)*pv++;
				case 3: h.pos = (unsigned short)pv[2];
				case 2: h.occ = (unsigned short)pv[1];
				case 1: h.tag = (unsigned short)pv[0];
				/* case 0: ! f->len */
				}
				if ( 4 > i && rec )
					h.mfn = rec->rowid;
			} break; /* case XDOT */
			case XFST:
				return log_msg( ERR_IDIOT, "sorry, XFST not implemented" );
			}
			if ( ! k.val.len )
				cXMkVal( &db->oxi, &k.val, &h );
			while ( f < last && !f[1].tag ) { /* collect XKEYs */
				int l = (++f)->len;
				unsigned char *dst = k.byt;
				const unsigned char *src = (const unsigned char *)f->val;
				if ( 255 < l )
					l = 255;
				k.len = (unsigned char)l;
				while ( l-- )
					*dst++ = uc[ *src++ ];
				LOG_DBG( LOG_DEBUG,
					"field %d key '%.*s' hit %d.%d.%d.%d.%d",
					f - idx->field, k.len, k.byt,
					h.dbn, h.mfn, h.tag, h.occ, h.pos );
				ret = del ? lbt_del( &db->oxi, &k ) : lbt_add( &db->oxi, &k );
			}
		}
	}
	return ret;
}	/* dWritex */


int dWrite ( Session ses, int dbid, Rec *rec )
{
	/* TODO: use FST lines as idx */
	return dWritex( ses, dbid, rec, 0 );
}	/* dWrite */


Rec* dTerm ( Session ses, Rec *rec, int dbid, const char *key )
{
	(void)ses;
	return 0 > ldb_search( dbid, key, 0, rec ) ? 0 : rec;
}	/* dTerm */


int dXLoop ( Session ses, int dbid, DXLoop *l )
{
	LDb *db = getDb( dbid );
	(void)ses;
	if ( !db )
		return -ERR_BADF;
	if ( OPENISIS_IDXTRAD & l->flg ) {
		if ( !(db->flags & DB_INVOPEN) )
			return -ERR_BUSY;
		return search( db, 0, 0, 0, l );
	}
	if ( !(db->flags & DB_LBTOPEN) || db->oxi.bat )
		return -ERR_BUSY;
	return lbt_loop( & db->oxi, l );
}	/* dXLoop */



int cInit ( int argc, const char **argv, CLockFunc lockfunc )
{
	(void)argc; (void)argv;
	cOpen( 0 );
	if ( lockfunc )
		lio_lock = lockfunc;
	return 0;
}


Db* cDOpen (const char *dbname, Rec *dbpar, Rec *syspar) {
	int dbid;
	if ( ! init )
		cOpen( 0 );
	dbid = ldb_open (dbname, dbpar, syspar);
	if (0 <= dbid) {
		return &dbs[dbid].head;
	}
	return 0;
}

int cDOpenv ( const char *dbname, const char **argv, int argc )
{
	Rec *parms = 0;
	int  rt;
	if ( ! init )
		cOpen( 0 );
	if (argc) {
		if (argc & ~RARGC) {
			log_msg (LOG_WARN, "illegal argc %d for dbopenv %s",
				argc, dbname ? dbname : "");
		}
		else {
			parms = rSet (ses0, 0,
				RARGV | RFDT | RNOC | RIGN | argc,
				openIsisSyspar, argv);
		}
	}
	rt = ldb_open (dbname, parms, 0);
	if (parms) {
		mFree (ses0, parms);
	}
	return rt;
}


int cDClose ( int dbid )
{
	LDb *db = getDb( dbid );
	if ( ! db )
		return -ERR_BADF;
	if ( LIO_LOCK() ) return -ERR_BUSY;
	if ( DB_MODIFIED & db->flags ) {
		/* write back the MF control */
		Mfc mfc;
		/* if ( 498 < (db->mflen & 511) ) db->mflen = ~511 & (db->mflen + 14); */
		mfc.ctlm = rvi( db->mfc[LMFC_CTLM] );
		mfc.nmfn = rvi( db->mfc[LMFC_NMFN] );
		mfc.nmfb = rvi( 1 + (db->mflen >> 9) );
		mfc.nmfp = rvs( 511 & db->mflen );
		mfc.type = rvs( db->mfc[LMFC_TYPE] );
		mfc.rcnt = rvi( db->mfc[LMFC_RCNT] );
		mfc.mfx1 = rvi( db->mfc[LMFC_MFX1] );
		mfc.mfx2 = rvi( db->mfc[LMFC_MFX2] );
		mfc.mfx3 = rvi( db->mfc[LMFC_MFX3] );
		if ( sizeof(mfc) != lio_pwrite( &db->mst[MST_MST], &mfc, sizeof(mfc), 0) )
			log_msg( ERR_TRASH, "could not write MST header" );
	}
	if ( db->mmap )
		lio_mmap( 0, (void**)&db->mmap, db->mmlen*512 );
	db->mmlen = 0;
	closefiles( db->mst, MST_FILES );
	if ( DB_INVOPEN & db->flags )
		closefiles( db->inv, INV_FILES );
	db->flags = 0L;
	if ( db->path ) free( (char*)db->path );
	if (db->head.cfg) mFree (ses0, db->head.cfg);
	if (db->head.fdt) cFreeFdt (ses0, db->head.fdt);
	memset( db, 0, sizeof(db) );
	(void)LIO_RELE();
	return 0;
}	/* cDClose */


int cDCheck ( int dbid, int flags )
{
	static char dot = '.';
	int *r;
	LDb *db = getDb( dbid );
	int nxtoff = 64, off;

	if ( ! db )
		return -ERR_BADF;
	(void)flags;
	do {
		lio_write( &lio_out, &dot, 1 );
		if ( (r = ldb_readRecAtOff(dbid,off=nxtoff,&nxtoff)) ) {
			int o = getOff( db, r[LMFR_MFN], 0 );
			if ( o != off ) {
				log_msg( LOG_WARN, "mfn %d xrf %d != real %d\n",
					r[LMFR_MFN], o, off );
			}
			free( r );
		}
	} while ( 0 < nxtoff );
	return 0;
}	/* cDCheck */


OpenIsisIdx *cXOpen ( int dbid, int mode )
{
	LDb *db = getDb( dbid );
	if ( !db
		|| !(db->flags & DB_LBTOPEN)
		|| !(db->oxi.flg & LBT_WRITE) /* may be writable if db is not */
		|| db->oxi.bat
		/*
		preliminary undocumented feature:
		mode -1 gives direct access in non-batch mode
		*/
		|| (0 <= mode && lbt_batch( & db->oxi, (unsigned char)mode ))
	)
		return 0;
	return & db->oxi;
}	/* cXOpen */
