/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: lses.c,v 1.12 2003/03/24 21:03:54 kripke Exp $
	implementation of general db access functions.
*/

#include <stdlib.h>
#include <string.h> /* memset */

#define LSES_C
#include "lses.h"


/* ************************************************************
	private types
*/
typedef struct {
	Ses h;
	int          brk;
} SesI;

enum {
	LSES_MAX = 0x1000,
	LSES_MSK = 0x0fff
};

/* ************************************************************
	package data
*/

/* ************************************************************
	private data
*/
static SesI lses0; /* the default session */
static SesI *tab[ LSES_MAX ] = {
	&lses0
};
static Ses *htab[ LSES_MAX ];
static int maxid;

/* ************************************************************
	private functions
*/
void lses_tini ()
{
	int i, s = LSES_MAX;
	/* lio_write( &lio_out, "tini time\n", 9 ); */
	while ( s-- )
		if ( tab[s] )
			for ( i=0; i<3; i++ )
				if ( tab[s]->h.io[i] )
					LIO_CLOSE( tab[s]->h.io[i] );
}	/* lses_tini */


/* ************************************************************
	package functions
*/

/* legacy */
void * lmalloc ( int siz )
{
	static const char oops[] = "out of memory\n";
	void *mem = malloc( siz );
	if ( ! mem )
		lio_write( &lio_err, oops, sizeof(oops)-1 );
	return mem;
}	/* lmalloc */


/* ************************************************************
	public functions
*/
Session ses0;

Ses *cOpen ( Rec *args )
{
	if ( ! ses0 ) {
		ses0 = &lses0.h;
		lio_init();
		lses0.brk = 1; /* fake */
		sOpen ( ses0, "<&0", 0, lio_stdio );
		sOpen ( ses0, ">&1", 0, lio_stdio );
		sOpen ( ses0, "2>&2", 0, lio_stdio );
		atexit( lses_tini );
	}
	if ( args ) { /* add args */
		/* TODO: merge args */
		lses0.h.prop = rDup( ses0, args, -1, 0 );
	}
	return ses0;
}	/* cOpen */


Ses *cSession ( Rec *args )
{
	SesI *sesi = 0;
	Ses *ses = 0;
	int id = 1;
	/** XXX unless we actually free sessions,
		we should start checking with maxid
		or check for expiry
	*/
	while ( tab[id] )
		if ( LSES_MAX == ++id )
			return 0;
	if ( maxid < id )
		maxid = id;
	sesi = tab[id] = (SesI*)malloc( sizeof(SesI) );
	memset( sesi, 0, sizeof(SesI) );
	ses = &sesi->h;
	ses->id = id;
	ses->hash = -1; /* not hashed (hash is non-negative) */
	/* prepare session heap */
	sesi->brk = 1; /* fake */
	ses->prop = args ? rDup( ses, args, -1, 0 ) : 0;
	/* check props */
	/* open files */
	sOpen ( ses, "2>&2", 0, lio_stdio );
	return ses;
}	/* cSession */


void *mAlloc ( Session s, int size )
{
	void *p = malloc( size );
	(void)s;
	if (p) {
		memset (p, 0, size);
	}
	return p;
}	/* mAlloc */


void  mFree ( Session s, void *mem )
{
	(void)s;
	if ( mem )
		free( mem );
}	/* mFree */


void *mDup ( Session s, const void *str, int sz )
{
	void *m = mAlloc( s, 0<=sz ? sz : (sz = strlen( str ) + 1) );
	if ( m )
		memcpy( m, str, sz );
	return m;
}	/* mDup */


int sOpen ( Session s, const char *name, int flags,
	SFunc *type )
{
	int fd = 0;
	const char *p = name;
	lio_sfunc *func = type ? (lio_sfunc*)type : lio_stdio;

	/* check for i<> */
	while ( '0' <= *p && *p <= '9' )
		fd = 10*fd + *p++ - '0';
	if ( '<' != *p && '>' != *p )
		p = name; /* forget it */
	else {
		if ( name == p && '>' == *p ) /* no fd given, > defaults to 1 */
			fd = 1;
		flags |= '<' == *p ? LIO_RD : LIO_WR;
		p++;
	}
	if ( p == name ) { /* no fd given, find one */
		for ( fd=0; s->io[fd]; )
			if ( LSES_FILE_MAX == ++fd )
				return -ERR_NOMEM;
	} else if ( s->io[fd] ) { /* close existing */
		func( s->io[fd], LIO_SCLOSE );
		mFree( s, s->io[fd] );
	}
	s->io[fd] = (Ios*)mAlloc( s, func( 0, LIO_SSIZE ) );
	if ( ! s->io[fd] )
		return -ERR_NOMEM;
	s->io[fd]->func = func;
	s->io[fd]->name = mDup( s, p, -1 );
	s->io[fd]->file = flags;
	return func( s->io[fd], LIO_SOPEN );
}	/* sOpen */


Ses *openIsisId2Session (int sid) {
	if (0 > sid || LSES_MAX <= sid || 0 == tab[sid]) {
		return (Ses*)0;
	}
	return &(tab[sid]->h);
}	/* openIsisId2Session */


int openIsisSession2Id (Ses *ses) {
	if (0 == ses) {
		return -1;
	}
	return ses->id;
}	/* openIsisSession2Id */


Ses *cSesByName ( char *name, int nlen, Tm *now, Tm *expire )
{
	char tbuf[20];
	int h,i;
	Ses *s;
	if ( nlen < 0 )
		nlen = strlen(name);
	if ( nlen > 63 )
		nlen = 63;
	for ( i=nlen; i--; )
		if ( !(0xe0 & name[i]) )
			name[i] = '_';
	h = lhash( name, nlen );
	for ( s = htab[h & LSES_MSK]; s; s = s->nxt )
		if ( h == s->hash
			&& !memcmp( name, s->name, nlen )
			&& !s->name[nlen]
		) {	/* match */
			if ( !expire || s->atime.millis >= expire->millis )
				goto gotit; /* valid */
			if ( s->cur || s->que ) {
				log_msg( LOG_WARN, "expired session %d %s still busy!",
					s->id, s->name );
				goto gotit; /* valid anyway */
			}
			goto refresh;
		}
	/* didn't find by name */
	if ( ! expire ) /* bad thing -- we're going to run out of sessions ... */
		i = maxid+1;
	else /* check other expired */
		for ( i=1; i<=maxid; i++ ) {
			s = &tab[i]->h; /* sind wir heute defensiv ... */
			if ( s && s->atime.millis < expire->millis
				&& s->atime.millis && 0 <= s->hash /* else is none of our biz */
				&& !s->cur && !s->que /* else is busy !? */
			)
				break;
		}
	if ( i > maxid ) {
		s = cSession( 0 );
		if ( ! s )
			return 0;
	} else /* got one */
		log_msg( LOG_INFO, "reusing session %d '%s' atime %s",
			s->id, s->name, s->atime.millis ? timeGtf(tbuf,&s->atime) : "-" );
	if ( 0 <= s->hash ) { /* was hashed -- unlink */
		Ses **spp = &htab[s->hash & LSES_MSK];
		for ( ; *spp; spp = &(*spp)->nxt )
			if ( s == *spp ) {
				*spp = s->nxt;
				break;
			}
	}
	memset( s->name, 0, sizeof(s->name) );
	memcpy( s->name, name, nlen );
	s->hash = h;
	s->nxt = htab[h & LSES_MSK];
	htab[h & LSES_MSK] = s;
refresh:
	s->ctime.millis = s->mtime.millis = now ? now->millis : 0;
	s->accnt = 0;
	if ( s->prop )
		CLRREC( s->prop );
	if ( s->res )
		CLRREC( s->res );
gotit:
	s->atime.millis = now ? now->millis : 0;
	return s;
}	/* cSesByName */
