/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: lstb.c,v 1.21 2003/03/24 14:34:38 kripke Exp $
	implementation of client side session and communication functions.
*/

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "openisis.h"
#include "loi.h"
#include "lcom.h"
#include "ldb.h"
#include "lfdt.h"
#include "lses.h"
#include "luti.h"

#ifdef WIN32
#define snprintf _snprintf
#define vsnprintf _vsnprintf
#endif

#define LSTB_STHOSTLEN   32

#define LSTB_MAXDBS      255  /* according to OPENISIS_DBIDMSK */
#define LSTB_MAXSTUBS    255  /* according to OPENISIS_SCIDMSK */

#define LSTB_INCRDB      8
#define LSTB_INCRSTUB    8

/* ************************************************************
 * meta
*/

static Fd _RqsFields[] = {
/* tag type rep len name subf desc */
	{ LCOM_RQS_SID,   FTN, 0, 2, "sid", "", "Client Session Id" },
	{ LCOM_RQS_SER,   FTN, 0, 5, "ser", "", "Request Serial No." },
	{ LCOM_RQS_TYPE,  FTE, 0, 32, "type", "", "Client Session Id" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_OPEN, "open", "", "open db" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_CREA, "create", "", "create db" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_CLOS, "close", "", "close db" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_MROW, "maxrow", "", "get maxrowid" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_QRY,  "query", "", "exec query" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_READ, "read", "", "fetch row" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_INS,  "insert", "", "insert rec" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_UPD,  "update", "", "update rec" },
	{ LCOM_RQS_TYPE,  FTV, 0, LCOM_RQST_DEL,  "delete", "", "delete row" },
	{ LCOM_RQS_DBN,   FTX, 0, DBNLEN, "db", "", "DB Identification" },
	{ LCOM_RQS_TMS,   FTN, 0, 5, "tms", "", "Server Db Timestamp" },
	{ LCOM_RQS_ROWID, FTN, 0, 5, "rowid", "", "RowId" },
	{ LCOM_RQS_QMOD,  FTN, 0, 5, "mode", "", "Query Mode" },
	{ LCOM_RQS_SKIP,  FTN, 0, 5, "skip", "", "Query Skip" },
	{ LCOM_RQS_SIZE,  FTN, 0, 5, "size", "", "Result Length" },
	{ LCOM_RQS_QFLG,  FTN, 0, 5, "flags", "", "Query Flags" },
	{ LCOM_RQS_KEY,   FTX, 0, LCOM_QRY_KEYLEN, "key", "", "Query Key" },
	{ LCOM_RQS_REC,   FTS, 1, 0, "rec", "", "Record to Update" },
	{ LCOM_RQS_IDX,   FTS, 1, 0, "idx", "", "Index to Update" }
};

static Fdt _RqsFdt = {
	sizeof(_RqsFields) / sizeof(_RqsFields[0]),
	_RqsFields,
	0
};

static Fd _RspFields[] = {
/* tag type rep len name subf desc */
	{ LCOM_RSP_SID,   FTN, 0, 2, "sid", "", "Client Session Id" },
	{ LCOM_RSP_SER,   FTN, 0, 5, "ser", "", "Request Serial No." },
	{ LCOM_RSP_DBN,   FTX, 0, DBNLEN, "db", "", "DB Identification" },
	{ LCOM_RSP_TMS,   FTN, 0, 5, "tms", "", "Server Db Timestamp" },
	{ LCOM_RSP_ERR,   FTN, 0, 5, "error", "", "Error Code" },
	{ LCOM_RSP_CERR,  FTN, 0, 5, "error2", "", "Client Side Error" },
	{ LCOM_RSP_MSG,   FTX, 0, LCOM_ERRMSGLEN, "msg", "", "Error Message" },
	{ LCOM_RSP_NUMT,  FTN, 0, 5, "total", "", "Total No. of Records" },
	{ LCOM_RSP_NUMR,  FTN, 0, 5, "size", "", "Number of Records" },
	{ LCOM_RSP_ROWID, FTN, 1, 5, "rowid", "", "RowId" },
	{ LCOM_RSP_DATA,  FTS, 1, 0, "data", "", "Fetched Structure" }
};

static Fdt _RspFdt = {
	sizeof(_RspFields) / sizeof(_RspFields[0]),
	_RspFields,
	0
};

const Fdt *openIsisComRqsPar = &_RqsFdt;
const Fdt *openIsisComRspPar = &_RspFdt;

/* stub id 0      has idx -1 */
/* stub id 0x0100 has idx  0 */

#define SC_X2I(s)  ( ( (1 + (s)) << 8 ) & OPENISIS_SCIDMSK )

#define SC_I2X(s)  ( ( ((s) & OPENISIS_SCIDMSK) >> 8 ) - 1 )

#define DB_X2I(d,s)  ( ((s) & OPENISIS_SCIDMSK) | ((d) & OPENISIS_DBIDMSK) )

#define DB_I2X(d)    ( (d) & OPENISIS_DBIDMSK )

#define DB_IDX_OK(st,idx)   \
	(0 <= (idx) && (st)->head.ndbs > (idx) && * (st)->head.dbs[idx]->name)

/* ************************************************************
 * client view of db
*/

typedef struct {
	Db head;
} CDb;

static void cdbCtor (CDb *that,
	const char *name, int dbid, int tms, Fdt *fdt
) {
	strncpy (that->head.name, name, DBNLEN - 1) [DBNLEN - 1] = 0;
	that->head.dbid = dbid;
	that->head.tms  = tms;
	that->head.fdt  = fdt;
}

static void cdbDtor (CDb *that) {
	if (that->head.fdt) {
		mFree (ses0, that->head.fdt);
	}
	memset (that, 0, sizeof (CDb));
}

/* ************************************************************
 * client session
*/

typedef struct OpenIsisStb {
	Schema    head;
	char      host[LSTB_STHOSTLEN];
	int       port;
	int       srq;   /* rqs serial no */
	Rec      *rqs;   /* last request */
	Rec      *rsp;   /* last response */
	Db       *rdb;   /* db of embedded rsp recs */
	Session   ses;   /* session used for actual/last rqs/rsp */
	LutiLT    hsh;
	CDb      *dbase; /* db array of non-local schema */
	OpenIsisRspCb       *actcb;
	void                *actcld;
	OpenIsisStubCbData   cbdta;
} OpenIsisStb;

Stub stub0 = 0;

static Stub  RmtStubs = 0;
static int   NumStubs = 0;

static LutiLT StubLT = 0;

#define STBSIZE  (sizeof(Stb))

static void stbCtor (Stub that,
	const char *name, const char *host, int port, int scid,
	OpenIsisStubCbData *cbd
) {
	strncpy (that->head.name, name, SCNLEN - 1) [SCNLEN - 1] = 0;
	strncpy (that->host, host, LSTB_STHOSTLEN - 1) [LSTB_STHOSTLEN - 1] = 0;
	that->head.scid = scid;
	that->port = port;
	that->rdb  = 0;
	that->hsh  = luti_ltnew ();
	if (cbd) {
		memcpy (&that->cbdta, cbd, sizeof (OpenIsisStubCbData));
	}
	luti_ltadd (StubLT, that->head.name, scid);
}

static int cdbNewIdx (Stub that) {
	char  **D, **E;
	CDb    *obase;
	int     ndb = that->head.ndbs;
	int     rt, j, diff;
	for (j = 0; ndb > j; ++j) {
		if (! * that->dbase[j].head.name) {
			return j;
		}
	}
	obase = that->dbase;
	rt = luti_ptrincr (ses0, &that->dbase, &ndb,
		LSTB_INCRDB, sizeof (CDb), LSTB_MAXDBS);
	if (0 > rt) {
		return -1;
	}
	rt = luti_ptrincr (ses0, &that->head.dbs, &that->head.ndbs,
		LSTB_INCRDB, sizeof (CDb*), LSTB_MAXDBS);
	if (0 > rt) {
		return -1;
	}
	diff = (char*) that->dbase - (char*) obase;
	if (diff) {
		E = (D = (char**) that->head.dbs) + rt;
		while (E > D) {
			*D += diff;
			++D;
		}
	}
	return rt;
}

static int cdbNewIdx0 (Stub that, int dbid, int *idx) {
	int  rt;
	*idx = dbid;
	if (dbid < that->head.ndbs) {
		if (that->head.dbs[dbid]) {
			return 0;
		}
		return 1;
	}
	rt = luti_ptrincr (ses0, &that->head.dbs, &that->head.ndbs,
		1 + dbid - that->head.ndbs, sizeof (CDb*), -1);
	if (0 > rt) {
		return -1;
	}
	return 1;
}

static int stbNewIdx () {
	int   j;
	for (j = 0; NumStubs > j; ++j) {
		if (! * RmtStubs[j].head.name) {
			return j;
		}
	}
	return luti_ptrincr (ses0, &RmtStubs, &NumStubs,
		LSTB_INCRSTUB, STBSIZE, LSTB_MAXSTUBS);
}

static Stub stbById (int sid) {
	int idx;
	if (0 > sid || ! stub0) {
		return 0;
	}
	if (0 == sid) {
		return stub0;
	}
	idx = SC_I2X (sid);
	if (0 > idx) {
		log_msg (LOG_ERROR, "stbById: illegal id %x\n", sid);
		return 0;
	}
	if (NumStubs <= idx) {
		return 0;
	}
	if (! * RmtStubs[idx].head.name) {
		return 0;
	}
	return RmtStubs + idx;
}

static void stbDtor (Stub that) {
	luti_ltrmv (StubLT, that->head.name);
	if (that->cbdta.delcb) {
		if (that->actcld) {
			(*that->cbdta.delcb) (
				that->cbdta.delcld, that, that->actcld);
		}
		if (that->cbdta.dfltcld) {
			(*that->cbdta.delcb) (
				that->cbdta.delcld, that, that->cbdta.dfltcld);
		}
		(*that->cbdta.delcb) (that->cbdta.delcld, that, 0);
	}
	if (that->dbase) {
		int j;
		for (j = that->head.ndbs; 0 <= --j;  ) {
			if (* that->dbase[j].head.name) {
				cdbDtor (that->dbase + j);
			}
		}
		mFree (ses0, that->dbase);
	}
	if (that->head.dbs) {
		mFree (ses0, that->head.dbs);
	}
	if (that->head.cfg) {
		mFree (ses0, that->head.cfg);
	}
	luti_ltdel (that->hsh);
	nClean (that);
	memset (that, 0, STBSIZE);
}

static Rec* stbMetaFor (const char *scname) {
	Rec   *ent = 0;
	Field *F   = 0;
	Field *N   = 0;
	int    pos = 0;

	if (! stub0->head.cfg) {
		return 0;
	}

	while (1) {
		F = rGet (stub0->head.cfg, OPENISIS_SSCHEMA, &pos);
		if (ent) {
			mFree (ses0, ent);
			ent = 0;
		}
		if (! F) {
			break;
		}
		ent = rSplitf (ses0, 0, F);
		if (! ent) {
			break;
		}
		N = rGet (ent, OPENISIS_SSCNAME, 0);
		if (! N) {
			log_msg (LOG_WARN, "illegal schema entry %d: %.*s\n",
				pos, (int)F->len, F->val);
			continue;
		}
		if (! strncmp (scname, N->val, (int)N->len)) {
			break;
		}
	}

	return ent;
}

static int stbDbidFromName (Stub that, const char *dbn) {
	int id;
	if (! dbn) {
		return -1;
	}
	id = luti_ltget (that->hsh, dbn);
	return id;
}

static int stbDbIdxFromName (Stub that, const char *dbn) {
	int idx = stbDbidFromName (that, dbn);
	if (0 > idx) {
		return -1;
	}
	idx = DB_I2X (idx);
	if (! DB_IDX_OK (that, idx)) {
		return -1;
	}
	return idx;
}

static int stbDbIdxFromRec (Stub that, Rec *rec, int tag) {
	char dbn[DBNLEN];
	return stbDbIdxFromName (that,
		rString (rec, tag, 0, dbn, DBNLEN));
}

static int stbDbidFromRec (Stub that, Rec *rec, int tag) {
	char dbn[DBNLEN];
	return stbDbidFromName (that,
		rString (rec, tag, 0, dbn, DBNLEN));
}

static int stbOpenDb (Stub that) {
	char  dbn[DBNLEN];
	Fdt  *fdt;
	int   idx, dbid, tms;
	int   newdb = 0;

	if (! rString (that->rsp, LCOM_RSP_DBN, 0, dbn, DBNLEN)) {
		log_msg (LOG_ERROR, "stbOpenDb: missing db name in rsp\n");
		return -1;
	}
	idx = stbDbIdxFromName (that, dbn);

	if (0 > idx) {
		idx = cdbNewIdx (that);
		if (0 > idx) {
			return -1;
		}
		newdb = !0;
	}
	else {
		cdbDtor (that->dbase + idx);
	}

	fdt = rec2fdt (ses0, that->rsp, LCOM_RSP_NUMR, LCOM_RSP_DATA, 0);
	tms = rInt (that->rsp, LCOM_RSP_TMS, 0, 0);
	dbid = DB_X2I (idx, that->head.scid);

	cdbCtor (that->dbase + idx, dbn, dbid, tms, fdt);

	if (newdb) {
		luti_ltadd (that->hsh, dbn, dbid);
	}
	return 0;
}

static int stbOpenDb0 (Stub that) {
	Db  *dbh;
	int  idx;
	int  newdb = 0;
	int  dbid  = rInt (that->rsp, LCOM_RSP_DBID, -1, 0);
	if (0 > dbid) {
		log_msg (LOG_ERROR, "stbOpenDb0: missing dbid in rsp");
		return -1;
	}

	newdb = cdbNewIdx0 (that, dbid, &idx);
	if (0 > newdb) {
		log_msg (LOG_ERROR, "stbOpenDb0: out of memory");
		return -1;
	}

	dbh = ldb_getdb (idx);
	if (! dbh) {
		log_msg (LOG_ERROR, "stbOpenDb0: no such db %d %d", dbid, idx);
		return -1;
	}
	log_msg (LOG_INFO, "stbOpenDb0 %d %d %s", dbid, idx, dbh->name);

	that->head.dbs[idx] = dbh;
	if (newdb) {
		luti_ltadd (that->hsh, dbh->name, idx);
	}
	return 0;
}

static int stbCloseDb (Stub that) {
	CDb *db;
	int  dbid = stbDbIdxFromRec (that, that->rsp, LCOM_RSP_DBN);
	if (0 > dbid) {
		log_msg (LOG_ERROR, "stbCloseDb: missing or illegal db name in rsp");
		return -1;
	}
	db = that->dbase + dbid;
	luti_ltrmv (that->hsh, db->head.name);
	cdbDtor (db);
	return 0;
}

static int stbCloseDb0 (Stub that) {
	char   dbn[DBNLEN];
	int    dbid = rInt (that->rsp, LCOM_RSP_DBID, -1, 0);
	if (0 > dbid) {
		log_msg (LOG_ERROR, "stbCloseDb0: missing dbid in rsp");
		return -1;
	}
	if (that->head.ndbs <= dbid) {
		log_msg (LOG_ERROR, "stbCloseDb0: illegal dbid %d(%d) in rsp",
			dbid, that->head.ndbs);
		return -1;
	}
	if (! rString (that->rsp, LCOM_RSP_DBN, 0, dbn, DBNLEN)) {
		log_msg (LOG_ERROR, "stbCloseDb0: missing dbn in rsp");
		return -1;
	}
	log_msg (LOG_INFO, "stbCloseDb0 %d %s", dbid, dbn);
	luti_ltrmv (that->hsh, dbn);
	that->head.dbs[dbid] = 0;
	return 0;
}

/*	(local) response callback
	MMM TBD if called from file event handler,
	read stub id, determine ses from stub,
	deserialize and allocate rsp and call cb
	NOTE: rsp may be discarded in this call (s. err2)
*/
static int stbRspCb (Session ois, Rec **rsp, void *cld) {
	Stub      that;
	Rec      *rsp2;
	int       sid, ser, err, err2, err3;

	(void)cld;
	sid = rInt (*rsp, LCOM_RSP_SID, -1, 0);
	ser = rInt (*rsp, LCOM_RSP_SER, -1, 0);
	err = rInt (*rsp, LCOM_RSP_ERR, -1, 0);

	that = stbById (sid);
	if (! that) {
		log_msg (LOG_WARN, "illegal stub id %d in stbRspCb", sid);
		return -1;
	}
	if (ser != that->srq) {
		log_msg (LOG_WARN, "serial mismatch %d %d stub %d",
			ser, that->srq, sid);
		return -1;
	}
	if (! that->rqs) {
		log_msg (LOG_WARN, "response without request %d stub %d",
			ser, sid);
		return -1;
	}

	if (ois != that->ses) {
		log_msg (LOG_ERROR, "session mismatch: got %d(%d), exp %d(%d)",
			ois, ois->id, that->ses, (that->ses ? that->ses->id : 0));
		return -1;
	}

	if (that->rsp) {
		log_msg (LOG_WARN, "multiple responses for request %d stub %d",
			ser, sid);
		mFree (that->ses, that->rsp);
	}
	that->rdb = 0;
	that->rsp = *rsp;
	*rsp = 0;

	err2 = err3 = 0;

	if (0 == err) {
		int rtyp = rInt (that->rqs, LCOM_RQS_TYPE, -1, 0);
		if (LCOM_RQST_CLOS == rtyp) {
			err2 = sid ? stbCloseDb (that) : stbCloseDb0 (that);
		}
		else {
			int rdid, rx;
			if (LCOM_RQST_OPEN == rtyp || LCOM_RQST_CREA == rtyp) {
				err2 = sid ? stbOpenDb (that) : stbOpenDb0 (that);
			}
			rdid = stbDbidFromRec (that, that->rsp, LCOM_RSP_DBN);
			if (0 > rdid) {
				log_msg (LOG_ERROR, "stbRspCb: unknown db id\n");
				err2 = -1;
			}
			/* see nDbById */
			rx = DB_I2X (rdid);
			if (! DB_IDX_OK (that, rx)) {
				log_msg (LOG_ERROR, "stbRspCb: illegal db id %x\n", rdid);
				err2 = -1;
			}
			else {
				that->rdb = that->head.dbs[rx];
			}
		}
	}
	else {
		int tms;
		tms = rInt (that->rsp, LCOM_RSP_TMS, 0, 0);
		if (tms && sid) { /* remounted */
			err2 = stbCloseDb (that);
		}
	}

	rsp2 = rAddI (that->ses, that->rsp, LCOM_RSP_CERR, err2, !0);
	if (rsp2) {
		that->rsp = rsp2;
	}

	if (that->actcb) {
		err3 = (*that->actcb) (that->actcld,
			that, that->rsp, that->rdb);
		if (that->actcld && that->cbdta.delcb) {
			(*that->cbdta.delcb) (
				that->cbdta.delcld, that, that->actcld);
		}
		that->actcb = 0;
		that->actcld = 0;
	}
	else if (that->cbdta.dfltcb) {
		err3 = (*that->cbdta.dfltcb) (that->cbdta.dfltcld,
			that, that->rsp, that->rdb);
	}

	/*	if (err) return err;
		nSend should not return processing errors
	*/
	if (err2) return err2;
	return err3;
}

/* ************************************************************
	package functions
*/

/* ************************************************************
	public functions
*/

Stub openIsisNInit (
	int argc, const char **argv, OpenIsisStubCbData *dta
) {
	Session  ses;

	if (stub0) {
		return stub0;
	}

	ses = OPENISIS_SES0 ();
	StubLT = luti_ltnew ();
	stub0 = mAlloc (ses, STBSIZE);
	memset (stub0, 0, STBSIZE);
	stbCtor (stub0, "", "", 0, 0, dta);

	if (argc) {
		if (argc & ~RARGC) {
			log_msg (LOG_WARN, "illegal argc %d for stub0", argc);
		}
		else {
			stub0->head.cfg = rSet (ses, 0, RFDT | RARGV | RIGN | argc,
				openIsisSyspar, argv);
		}
	}

	return stub0;
}

Stub openIsisNOpen (
	const char *name, int argc, const char *argv[], OpenIsisStubCbData *dta
) {
	char  host[LSTB_STHOSTLEN];
	Stub  that  =  0;
	Rec  *meta  =  0;
	int   idx, port;

	if (! stub0 || ! name || ! *name) {
		return 0;
	}

	(void)argc; (void)argv; /* MMM TBD */

	idx = luti_ltget (StubLT, name);
	if (0 <= idx) {
		idx = SC_I2X (idx);
		if (0 > idx || NumStubs <= idx || ! * RmtStubs[idx].head.name) {
			log_msg (LOG_ERROR, "nOpen: illegal idx %d(%d) for %s\n",
				idx, NumStubs, name);
			return 0;
		}
		return RmtStubs + idx;
	}

	meta = stbMetaFor (name);
	if (! meta) {
		log_msg (LOG_WARN, "nOpen: unknown schema <%s>\n", name);
		goto done;
	}

	if (! rString (meta, OPENISIS_SSCHOST, 0, host, LSTB_STHOSTLEN)) {
		log_msg (LOG_WARN, "nOpen: no host for schema %s\n", name);
		goto done;
	}
	port = rInt (meta, OPENISIS_SSCPORT, 0, 0);
	if (! port) {
		log_msg (LOG_WARN, "nOpen: no port for schema %s\n", name);
		goto done;
	}

	idx = stbNewIdx ();
	if (0 > idx) {
		goto done;
	}
	that = RmtStubs + idx;
	idx = SC_X2I (idx);
	stbCtor (that, name, host, port, idx, dta);

done:
	if (meta) {
		mFree (ses0, meta);
	}
	return that;
}

void openIsisNClose (Stub that) {
	if (that && *that->head.name) {
		stbDtor (that);
	}
}

void openIsisNDeinit () {
	if (stub0) {
		int j;
		if (RmtStubs) {
			for (j = NumStubs; 0 <= --j;  ) {
				if (* RmtStubs[j].head.name) {
					stbDtor (RmtStubs + j);
				}
			}
			mFree (ses0, RmtStubs);
			RmtStubs = 0;
			NumStubs = 0;
		}
		for (j = stub0->head.ndbs; 0 <= --j;  ) {
			if (stub0->head.dbs[j]) {
				cDClose (stub0->head.dbs[j]->dbid);
			}
		}
		stbDtor (stub0);
		mFree (ses0, stub0);
		stub0 = 0;
		luti_ltdel (StubLT);
		StubLT = 0;
	}
}

Schema* nSchema (Stub that) {
	int idx;
	if (that && ( stub0 == that || ( *that->head.name &&
		0 <= (idx = SC_I2X (that->head.scid)) && NumStubs > idx
	))) {
		return &that->head;
	}
	return 0;
}

void nClean (Stub that) {
	if (that->ses) {
		if (that->rqs) {
			mFree (that->ses, that->rqs);
			that->rqs = 0;
		}
		if (that->rsp) {
			mFree (that->ses, that->rsp);
			that->rsp = 0;
		}
		that->rdb = 0;
		that->ses = 0;
	}
}

int nSend (Stub that,
	Session ses, Rec *rqs, OpenIsisRspCb *actcb, void *actcld, int dup
) {
	char  sid[16];
	char  ser[16];
	char  tms[16];
	int   dbid;

	if (!(that && ses && rqs)) {
		return -1;
	}
	if (stub0 != that && ! *that->head.name) {
		return -1;
	}
	if (that->ses) {
		if (! that->rsp) {
			log_msg (LOG_WARN, "stub %s: attempt to send multiple requests",
				that->head.name);
			return -1;
		}
		nClean (that);
	}

	i2a (sid, that->head.scid);
	if (0 >= ++that->srq) {
		that->srq = 1;
		log_msg (LOG_INFO, "stub %s: srq loop over 0", that->head.name);
	}
	i2a (ser, that->srq);
	dbid = -1;
	if (0 > rInt (rqs, LCOM_RQS_TMS, -1, 0)) {
		dbid = stbDbIdxFromRec (that, rqs, LCOM_RQS_DBN);
		if (0 <= dbid) {
			i2a (tms, that->head.dbs[dbid]->tms);
		}
	}

	if (dup) {
		rqs = rDup (ses, rqs, 0, 0);
		if (! rqs) {
			return -1;
		}
	}

	rqs = rSet (ses, rqs, RFDT | RDIS | RCHG, openIsisComRqsPar,
		LCOM_RQS_SID, sid, LCOM_RQS_SER, ser,
		0 <= dbid ? LCOM_RQS_TMS : 0, tms, 0);
	if (! rqs) {
		return -1;
	}

	that->ses = ses;
	that->rqs = rqs;

	if (that->actcld && that->cbdta.delcb) {
		/* ooups */
		(*that->cbdta.delcb) (
			that->cbdta.delcld, that, that->actcld);
	}
	if (actcb) {
		that->actcb = actcb;
		that->actcld = actcld;
	}
	else {
		that->actcb = 0;
		that->actcld = 0;
	}

	return ldsp_process (ses, that->rqs, &stbRspCb, 0);
}

Rec* nRecv (Stub that, Db **db) {
	if (that && (
		stub0 == that || *that->head.name
	)) {
		if (db) {
			*db = that->rdb;
		}
		return that->rsp;
	}
	return 0;
}

Db* nDbByName (Stub that, const char *dbname) {
	if (that && (
		stub0 == that || *that->head.name
	)) {
		int dbid = stbDbIdxFromName (that, dbname);
		if (0 <= dbid) {
			return that->head.dbs[dbid];
		}
	}
	return 0;
}

Db* nDbById (int id) {
	Stub that = stbById (id & OPENISIS_SCIDMSK);
	if (! that) {
		return 0;
	}
	id = DB_I2X (id);
	if (! DB_IDX_OK (that, id)) {
		return 0;
	}
	return that->head.dbs[id];
}

/* ************************************************************
	MMM util ? used in lstbtest
*/

#define FREEPTR(p)                                               \
	if (0 != p) {                                                \
		mFree (ses, p);                                          \
		p = 0;                                                   \
	}

int openIsisNGetResult (
	Stub that, Session ses, int **rowids, Rec ***recs, int *tot
) {
	Field     *F;
	Rec      **R;
	int       *I;
	int        numt, numr, err, pos, j;

	if (0 != rowids) {
		*rowids = 0;
	}
	if (0 != recs) {
		*recs = 0;
	}
	if (0 != tot) {
		*tot = 0;
	}
	if (! that || ! that->rsp || ! that->rdb || ! ses) {
		return 0;
	}
	err = rInt (that->rsp, LCOM_RSP_ERR, 0, 0);
	if (err) {
		return -1;
	}
	numr = rInt (that->rsp, LCOM_RSP_NUMR, -1, 0);
	numt = rInt (that->rsp, LCOM_RSP_NUMT, -1, 0);
	if (0 > numr || 0 > numt) {
		return -2;
	}
	if (numr) {
		if (rowids) {
			*rowids = (int*) mAlloc (ses, numr * sizeof (int));
			if (! *rowids) {
				return -3;
			}
			for (I = *rowids, j = pos = 0; numr > j; ++I, ++j) {
				*I = rInt (that->rsp, LCOM_RSP_ROWID, -1, &pos);
				if (0 > *I) {
					sMsg (ses, LOG_ERROR,
						"nGetResult: missing rowid %d(%d)", j, numr);
					FREEPTR (*rowids);
					return -2;
				}
			}
		}

		if (recs) {
			*recs = (Rec**) mAlloc (ses, numr * sizeof (Rec*));
			if (! *recs) {
				FREEPTR (*rowids);
				return -3;
			}
			memset (*recs, 0, numr * sizeof (Rec*));
			R = *recs;
			j = pos = 0;
			F = rGet (that->rsp, LCOM_RSP_DATA, &pos);
			if (! F) {
				sMsg (ses, LOG_WARN,
					"nGetResult: no records in response");
				FREEPTR (*recs);
			}
			else {
				--pos;
				while (numr > j) {
					*R = luti_unwrap (ses, that->rsp,
						&pos, LCOM_RSP_DATA, that->rdb->dbid);
					if (! *R) {
						FREEPTR (*rowids);
						luti_free (ses, (void**)*recs, j);
						*recs = 0;
						return -2;
					}
					++R;
					++j;
				}
			}
		} /* recs */

	} /* numr */

	if (0 != tot) {
		*tot = numt;
	}

	return numr;
}

