/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */


package org.openisis;

import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;


/**
	This class represents an isis db.
	The actual implementation is done in derived classes.
<p>
	$Id: Db.java,v 1.5 2003/03/05 19:52:28 istr Exp $
	@version $Revision: 1.5 $
	@author  $Author: istr $
*/
public abstract class Db {

	/** the default encoding.
	 *	the constant String value is "ISO-8859-1".
	*/
	public static final String DEFAULT_ENCODING = "ISO-8859-1";

	/* standard empty String returned by several methods */
	public static final String EMPTY = "";

	/** index scan key equal */
	public static final int QRY_KEYEQ = 0;
	/** index scan key prefix */
	public static final int QRY_KEYPF = 1;
	public static final int PREFIX = QRY_KEYPF; // legacy alias
	/** index scan key auto (checks for '$') */
	public static final int QRY_KEYAT = 2;
	/** fulltext scan for equal */
	public static final int QRY_SCANE = 64;
	/** fulltext scan for contains */
	public static final int QRY_SCANC = 65;
	/** query expression with simple left-to-right binding */
	public static final int QRY_SIMPLE = 128;
	/** query expression with proper binding using precedence, () */
	public static final int QRY_PROPER = 129;

	/** field v: heading mode */
	public static final int MHL = 0x10000;
	/** field v: data flag (no effect w/o heading) */
	public static final int MD  = 0x20000;
	/** field v: data mode (includes heading flag) */
	public static final int MDL = 0x30000;
	/** field v: uppercase mode flag */
	public static final int MXU = 0x40000;
	/** field v: uppercase heading mode */
	public static final int MHU = 0x50000;
	/** field v: uppercase data mode */
	public static final int MDU = 0x70000;
	/** field v: index formatting flag */
	public static final int MI  = 0x80000;

	/** field v: 16bit UNICODE HTML. HTML-entities are escaped. */
	public static final int HTU = 0x100000;
	/** field v: 8bit ISO-8859-1 HTML.
		HTML-entities and all chars greater than 255 are escaped.
	*/
	public static final int HTI = 0x300000;
	/** field v: 7bit ASCII HTML.
		HTML-entities and all chars greater than 127 are escaped.
	*/
	public static final int HTA = 0x700000;

	/** rec v: repeated prefix plus operator (not on 1st occ) */
	public static final int PP = 0x2000000;
	/** rec v: repeated suffix plus operator (not on last occ) */
	public static final int PS = 0x1000000;
	/** rec v: dummy presence selector Dn */
	public static final int DS = 0x4000000;
	/** rec v: dummy absence selector Nn */
	public static final int NS = 0x8000000;

	/** rec v: occurence range up to last occ */
	public static final int LOCC = 0xffff << 16;

	static final Map _dbs = new HashMap();


	/** open an isis database.
	The following options are supported as text-style parameters:
	<ul>
	<li>-db &lt;name&gt;<br>
		basename of database. same as param dbname.
	</li>
	<li>-dbpath &lt;path&gt;<br>
		path to database. usefull when opening secondary indexes
		or other files whose names are not based on the db basename.
	</li>
	<li>-v &lt;level&gt;<br>
		set verbosity level
	</li>
	<li>-encoding &lt;enc&gt;<br>
		select encoding used in db bytes. default is DEFAULT_ENCODING.
	</li>
	<li>-puredb<br>
		select pure java implementation.
		by default the NativeDb is used.
	</li>
	</ul>
	@param dbname basename of database.
		the dbhome arg, if given, will be prepended, and the ISIS file
		extensions appended to build the actual filename.
		Therefore, depending on the value of dbhome and location
		of files, this may need to include a path.
		db may be <code>null</code>, if argv includes a dbname arg.
	@param argv array of names and values.
		May be <code>null</code>.
		argv contains parameter names, which may be prefixed
		by a dash '-'. Depending on the actual parameter, the next
		string may or must contain a corresponding parameter value.
		An optional parameter value may be omitted, if at end of
		argv or the next name is prefixed with a dash.
		To avoid ambiguity, values never start with a dash.
		See above for a list of supported parameter names.
		Unrecognized parameter names are ignored.
	@return a new Db (unless some RuntimeException is thrown)
	@throws IoException especially FileNotFoundExceptiond
		and UnsupportedEncodingException
	*/
	public static synchronized Db open ( String dbname, String[] args )
		throws IOException
	{
		String enc = DEFAULT_ENCODING;
		boolean usePureDb = false;
		Db db = (Db) _dbs.get( dbname );
		if ( null != db )
			return db;
		if ( null != args )
			for ( int i=0; i<args.length; i++ )
				if ( "-encoding".equals( args[i] ) && i < args.length-1 ) {
					// if this Encoding gives UnsupportedEncodingException,
					// let it go off now
					byte b[] = new byte[1];
					b[0] = (byte)'A';
					new String( b, args[i+1] );
					enc =  args[i+1];
				} else if ( "-puredb".equals( args[i] ) ) {
					// usePureDb = true;
					System.err.println( "sorry, -puredb not yet implemented" );
				}
		db = // usePureDb ? (Db) new PureDb( enc, dbname, args ) : (Db)
			new NativeDb( enc, dbname, args );
		// System.err.println( "Db.open('"+dbname+"'): "+db );
		_dbs.put( dbname, db );
		return db;
	}	// open


	public static int[] AND ( int[] set, int[] nset ) {
		int[] aset = new int[ // minimum
			set.length < nset.length ? set.length : nset.length ];
		int a=0, s=0, n=0;
		while ( a<aset.length && s<set.length && n<nset.length ) {
			if ( set[s] == nset[n] ) {
				aset[a++] = set[s++];
				n++;
			} else if ( set[s] < nset[n] )
				s++;
			else
				n++;
		}
		set = new int[a];
		System.arraycopy( aset, 0, set, 0, set.length );
		return set;
	}	// AND


	//
	//	so far for the statics
	//

	/** encoding used for records in this db.
	 *	NOTE that String args are not converted according to this encoding,
	 *	but rather are passed as Java UTF-8, so you better stick to 7 bits.
	*/
	public final String encoding;

	protected Db ( String enc ) {
		encoding = enc;
	}


	/**
	read a cooked row.
	@param rowid the mfn
	@return a Rec or null, if row not available.
	*/
	public abstract Rec readRow ( int rowid );

	/**
	write a cooked row.
	@param row the Rec to be written
	@return 0 if write succeeded, else a negative error code
	*/
	public abstract short writeRow ( Rec rec );

	/**
	write a cooked row together with index information.
	@param row the Rec to be written
	@return 0 on success else a negative error code
	*/
	public abstract short writeXRow ( Rec rec, Rec idx );

	/**
		read the first row with rowid &gt;= given rowid
		and containing txt in subfield tag (any, if 0).
		@param rowid the mfn
		@return a Rec or null, if row not available.
	*/
	public abstract Rec scanRow ( int rowid, int tag, String txt );

	/**
		read the array of rowids matching key.
		@param key key to look for
		@param mode one of the QRY_ constants above.
			may be ored with a tag&lt;&lt;16 to limit search to that field
	*/
	public abstract int[] search ( String key, int mode );

	public abstract Iterator terms ( String prefix );

}	// Db
