/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

package org.openisis;

import java.io.*;
import java.net.Socket;
import java.util.ArrayList;

/**
	Serialization of fields by plain protocol.
<p>
	$Id: PlainSerializer.java,v 1.3 2003/03/31 15:43:36 mawag Exp $
	@version $Revision: 1.3 $
	@author  $Author: mawag $
*/
public class PlainSerializer {

	private static class ProtocolViolation extends RuntimeException {
		ProtocolViolation () {
			super ();
		}
		ProtocolViolation (String msg) {
			super (msg);
		}
		ProtocolViolation (byte arr[]) {
			super ("got <" + new String (arr) + ">");
		}
	}

	private static final byte[] NL  = new byte [] { (byte)'\n' };
	private static final byte[] TAB = new byte [] { (byte)'\t' };

	private static PrintStream _Log = System.err;

	private InputStream _in;
	private OutputStream _out;
	private boolean _dbg = false;

	/**	where to write the error messages to.
		default is stderr.
	*/
	public static void SetLogStream (PrintStream log) {
		if (null != log) {
			_Log = log;
		}
	}

	private void warn (String msg, Throwable ex) {
		_Log.println ("WARN [" + this + "] " + msg + ": " + ex);
	}

	private void debug (String rw, byte[] arr, int len) {
		_Log.println (rw + " <" + new String (arr, 0, len) + ">");
	}

	private void _write (byte[] arr) throws Exception {
		if (_dbg) {
			debug ("w", arr, arr.length);
		}
		_out.write (arr);
	}

	private int _read (byte[] arr) throws Exception {
		int rd = _in.read (arr);
		if (_dbg) {
			debug ("r", arr, rd);
		}
		return rd;
	}

	/**	write end of record to output stream.
		@return success
	*/
	public boolean flush () {
		try {
			_write (NL);
			return true;
		}
		catch (Exception ex) {
			warn ("flush", ex);
		}
		return false;
	}

	/**	serialize one field.
		@return success
	*/
	public boolean write (Field fld) {
		int    idx, lst, len;
		byte[] val;

		//	MMM TBD encoding

		if (null == fld) {
			return false;
		}
		try {
			val = Integer.toString (fld.tag).getBytes ();
			_write (val);
			_write (TAB);
			if (null != fld.val && 0 < (len = fld.val.length ())) {
				for (lst = idx = 0;
					0 <= (idx = fld.val.indexOf (NL[0], lst));
					lst = idx
				) {
					++idx;
					val = fld.val.substring (lst, idx).getBytes ();
					_write (val);
					_write (TAB);
				}
				if (lst < len) {
					val = fld.val.substring (lst).getBytes ();
					_write (val);
				}
			}
			_write (NL);
			return true;
		}
		catch (Exception ex) {
			warn ("serializing " + fld, ex);
		}
		return false;
	}

	/**	serialize fields.
		@return success
	*/
	public boolean write (Field[] flds) {
		if (null == flds) {
			return false;
		}
		for (int j = 0; flds.length > j; ++j) {
			if (! write (flds[j])) {
				return false;
			}
		}
		return flush ();
	}

	/** deserialize fields.
		@return array of fields read from input stream, null on error
	*/
	public Field[] read () {
		byte[]       buf = new byte [1024];
		StringBuffer str = new StringBuffer (1024);
		ArrayList    lst = new ArrayList ();
		Field        fld;
		int          rd, pos, tag, vs, val;
		boolean      rdt; // reading tag
		boolean      lnl; // last byte was NL
		boolean      neg; // tag negative

		//	MMM TBD encoding

		try {
			tag = 0;
			rdt = true;
			neg = false;
			lnl = false;
all:
			while (true) {
				rd = _read (buf);
				if (0 >= rd) {
					throw new ProtocolViolation ("no more bytes available");
				}
				vs = pos = 0;
				while (pos < rd) {
					if (rdt) {
						if (TAB[0] == buf[pos]) {
							rdt = false;
							vs = ++pos;
							continue;
						}
						if (0 == tag && '-' == buf[pos]) {
							neg = true;
							++pos;
							continue;
						}
						val = buf[pos] - '0';
						if (0 > val || 9 < val) {
							if (0 == pos &&
								NL[0] == buf[0] &&
								0 == lst.size ()) {
								warn ("", new ProtocolViolation (
									"empty response"));
								break;
							}
							throw new ProtocolViolation (buf);
						}
						tag = 10 * tag + val;
						++pos;
						continue;
					}
					if (lnl) {
						lnl = false;
						if (TAB[0] == buf[pos]) {
							str.append ("\n");
							vs = ++pos;
							continue;
						}
						fld = new Field (neg ? - tag : tag, str.toString ());
						lst.add (fld);
						if (NL[0] == buf[pos]) {
							if (pos != rd - 1) {
								throw new ProtocolViolation ("trailing bytes");
							}
							break all;
						}
						str.setLength (0);
						rdt = true;
						neg = false;
						tag = 0;
						continue;
					}
					if (NL[0] == buf[pos]) {
						if (pos > vs) {
							str.append (new String (buf, vs, pos - vs));
						}
						lnl = true;
					}
					++pos;
				}
				if (! rdt && rd > vs) {
					str.append (new String (buf, vs, rd - vs));
				}
			}

			return (Field[]) lst.toArray (new Field [lst.size ()]);
		}
		catch (Exception ex) {
			ex.printStackTrace (_Log);
			warn ("read", ex);
		}
		return null;
	}

	/** Constructor with read and write stream.
	*/
	public PlainSerializer (InputStream in, OutputStream out) {
		_in = in;
		_out = out;
	}

	/**	Constructor for a socket connection to port@host.
		@exception all exceptions a socket might throw
	*/
	public PlainSerializer (String hostname, int port) throws Exception {
		try {
			Socket sock = new Socket (hostname, port);
			_in = sock.getInputStream ();
			_out = sock.getOutputStream ();
		}
		catch (Exception ex) {
			warn ("(" + hostname + ';' + port + ")", ex);
			throw ex;
		}
	}

	/**	Close communication channels.
	*/
	public void close () {
		if (null != _in) {
			try {
				_in.close ();
			}
			catch (Exception ex) {
			}
			_in = null;
		}
		if (null != _out) {
			try {
				_out.close ();
			}
			catch (Exception ex) {
			}
			_out = null;
		}
	}

//	//////////////////////////////////////////////////////////////////////

	private class PSWriter extends Thread {
		Field[] _rqs;
		int _numr;
		PSWriter (Field[] rqs, int numr) {
			_rqs = rqs;
			_numr = numr;
		}
		public void run () {
			for (int j = 0; _numr > j; ++j) {
				if (! write (_rqs)) {
					warn ("error writing rqs #" + j, null);
					System.exit (1);
				}
				try {
					synchronized (this) {
						wait ();
					}
				}
				catch (Exception ex) {
					warn ("wait", ex);
				}
			}
		}
	}

	/**	Tester.
		<br/>
		usage: PlainSerializer [-d] [hostname [port [numrqs]]] | -p [numrqs]
	*/
	public static void main (String argv[]) {
		try {
			String   novel = "blah";
			String   host = "localhost";
			int      port = 8080;
			int      numr = 1;
			boolean  pipe = false;
			boolean  dbg  = false;
			int      j    = 0;
			int      len  = argv.length;

			if (0 < len && argv[0].startsWith ("-d")) {
				dbg = true;
				++j;
				--len;
			}
			switch (len) {
			case 3:
				numr = Integer.parseInt (argv[2 + j]);
			case 2:
				if (argv[j].startsWith ("-p")) {
					pipe = true;
					numr = Integer.parseInt (argv[1 + j]);
					break;
				}
				port = Integer.parseInt (argv[1 + j]);
			case 1:
				if (argv[j].startsWith ("-p")) {
					pipe = true;
				}
				else {
					host = argv[j];
				}
			case 0:
				break;
			default:
				throw new IllegalArgumentException ();
			}

			PlainSerializer ps;
			if (pipe) {
				PipedInputStream pi = new PipedInputStream ();
				PipedOutputStream po = new PipedOutputStream (pi);
				ps = new PlainSerializer (pi, po);
			}
			else {
				ps = new PlainSerializer (host, port);
			}
			ps._dbg = dbg;
			for (j = 0; 8 > j; ++j) {
				novel = novel + novel;
			}
			Field[] rqs = new Field [] {
				new Field (-2, "localhorst"),
				new Field (0, "nothing"),
				new Field (1, "one"),
				new Field (2, null),
				new Field (3, ""),
				new Field (4, novel),
				new Field (5, "\nfive\nfire\t"),
				new Field (6, "\t\n\t\n"),
				new Field (7, "seven"),
				new Field (8, "\n")
			};
			int k, kk;
			Field[] rsp;
			PSWriter writer = ps.new PSWriter (rqs, numr);
			writer.start ();

			for (j = 0; numr > j; ++j) {
				rsp = ps.read ();
				if (null == rsp) {
					ps.warn ("error reading rsp #" + j, null);
					System.exit (1);
				}
				for (k = kk = 0; rsp.length > k; ++k) {
					if (rqs.length <= kk) {
						ps.warn ("length mismatch rsp #" + j + ": " +
							rsp.length + " != " + rqs.length, null);
						System.exit (1);
					}
					if (0 > rsp[k].tag) {
						if (! rsp[k].equals (rqs[kk])) {
							//	skip additional server fields
							if (dbg) {
								_Log.println ("skipping server field " +
									rsp[k] + " ...");
							}
							continue;
						}
					}
					if (! rsp[k].equals (rqs[kk])) {
						ps.warn ("field #" + k + " rqs #" + j +
							" differs: " + rsp[k] + " != " + rqs[kk], null);
						System.exit (1);
					}
					++kk;
				}
				if (0 != j && 0 == (j % 100)) {
					System.out.println (j + " ...");
				}
				synchronized (writer) {
					writer.notify ();
				}
			}
			ps.close ();

			System.out.println ("ok.");
		}
		catch (Exception ex) {
			ex.printStackTrace (_Log);
			_Log.println ("usage: " + PlainSerializer.class.getName () +
				" [-d]" +
				" [<hostname> [<port> [<numrequests>]]] | -p [<numrequests>]");
			System.exit (1);
		}
	}
}

