/*
	openisis - an open implementation of the ISIS database
	Version 0.8.x (microversion see file Version)
	Copyright (C) 2001-2002 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	This software is dedicated to the memory of Eckart Dietrich.

	This software is inspired by (but contains no code of) the iAPI
	Copyright (C) 2000 by Robert Janusz, rj@jezuici.krakow.pl.
	See iAPI.txt for what it contains.
EOH */

/*
	$Id: openisistcl.087.c,v 1.1 2003/03/06 15:59:23 mawag Exp $
	tcl/tk binding
*/

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openisis.h>
#include <openisistcl.h>

#ifdef WIN32
#define snprintf _snprintf
#endif

/* ************************************************************
	private functions
*/

#define SESPRE "oises"
#define SPLEN  5

#define MAXVALLEN 4096 /* <= INT_MAX */

#define BUFLEN 32
#define PRNBUF(fmt,arg) \
	memset (buf, 0, BUFLEN); \
	snprintf (buf, BUFLEN - 1, fmt, arg);
#define PRNBUF2(fmt,a1,a2) \
	memset (buf, 0, BUFLEN); \
	snprintf (buf, BUFLEN - 1, fmt, a1, a2);

typedef struct {
	OpenIsisSession ses;
	long            idx;
	int             dbid;
	OpenIsisSet     ois;
} OIT_Set;

static int OIT_CmdSess _ANSI_ARGS_ (
	( OpenIsisSession ses, Tcl_Interp *ip, int argc, const char **argv ) );

static int OIT_CmdDb _ANSI_ARGS_ (
	( int *dbid, Tcl_Interp *ip, int argc, const char **argv ) );

static int OIT_FreeDb _ANSI_ARGS_ (
	( int *dbid, Tcl_Interp *ip, int argc, const char **argv ) );

static int OIT_CmdSet _ANSI_ARGS_ (
	( OIT_Set *set, Tcl_Interp *ip, int argc, const char **argv ) );

static int OIT_FreeSet _ANSI_ARGS_ (
	( OIT_Set *set, Tcl_Interp *ip, int argc, const char **argv ) );

static int scanLong (const char *str, long *res) {
	char *endp;
	if (0 == str) {
		return 0;
	}
	*res = strtol (str, &endp, 0);
	if (LONG_MAX == *res || LONG_MIN == *res) {
		return 0;
	}
	if (endp == str) {
		return 0;
	}
	return 1;
}

static int scanInt (const char *str, int *res) {
	long lval;
	if (! scanLong (str, &lval)) {
		return 0;
	}
	if (INT_MAX <= lval || INT_MIN >= lval) {
		return 0;
	}
	*res = (int) lval;
	return 1;
}

static OpenIsisSession scanSes (const char *str) {
	int sid;
	if (strncmp (SESPRE, str, SPLEN)) {
		return (OpenIsisSession)0;
	}
	if (! scanInt (str + SPLEN, &sid)) {
		return (OpenIsisSession)0;
	}
	return openIsisId2Session (sid);
}

static int resultSes (Tcl_Interp *ip, OpenIsisSession ses) {
	char buf[BUFLEN];
	int  rt;

	rt = openIsisSession2Id (ses);
	if (0 > rt) {
		PRNBUF ("%d", rt);
		Tcl_SetResult (ip, buf, TCL_VOLATILE);
		return TCL_ERROR;
	}

	PRNBUF2 ("%s%d", SESPRE, rt);
	Tcl_CreateCommand (ip, buf, (Tcl_CmdProc*)OIT_CmdSess,
		(ClientData)ses, (Tcl_CmdDeleteProc*)0);
	Tcl_SetResult (ip, buf, TCL_VOLATILE);
	return TCL_OK;
}

static int _SetCnt = 0; /* ??? */
static int NewSetCounterFor (OIT_Set *set) {
	return _SetCnt++;
}

static char *Rec2String (OpenIsisRec *rec, Tcl_DString *ds) {
	Tcl_DStringInit (ds);

	if (0 != rec) {
		OpenIsisField *fld;
		char val[MAXVALLEN];
		char buf[BUFLEN];
		unsigned len;
		for (fld = rec->field; rec->len > fld - rec->field; ++fld) {
			Tcl_DStringStartSublist (ds);
			PRNBUF ("%ld", fld->tag);
			Tcl_DStringAppendElement (ds, buf);
			if (0 != fld->len) {
				if (0 == (len = (unsigned) fld->len) || MAXVALLEN <= len) {
					fprintf (stderr, "truncation of %ld %ld %ld\n",
						rec->rowid, fld->tag, fld->len);
					len = MAXVALLEN - 1;
				}
				strncpy (val, fld->val, len) [len] = 0;
			}
			else {
				*val = 0;
			}
			Tcl_DStringAppendElement (ds, val);
			Tcl_DStringEndSublist (ds);
		}
	}

	return Tcl_DStringValue (ds);
}

static int OIT_ReadRec (
	Tcl_Interp *ip, OpenIsisSession ses, int dbid, int row
) {
	Tcl_DString  ds;
	OpenIsisRec *rec = openIsisDRead (ses, dbid, row);
	Tcl_SetResult (ip, Rec2String (rec, &ds), TCL_VOLATILE);
	if (0 != rec) {
		free (rec);
	}
	Tcl_DStringFree (&ds);
	return TCL_OK;
}

static int Append2Rec (
	OpenIsisSession ses, OpenIsisRec **recp, const char *tag, const char *val
) {
	int ntag;
	if (0 == val || ! scanInt (tag, &ntag) || 0 >= ntag) {
		return 0;
	}
	OPENISIS_RADDS (ses, *recp, ntag, val, !0);
	return 1;
}

/*	============================ general ================================
*/

/*	openIsisCLog ( int level, const char *filename )
*/
static int OIT_Log (
	ClientData cd, Tcl_Interp *ip, int argc, const char **argv
) {
	const char *fname = (char*)0;
	int lvl;

	switch (argc) {
	case 3:
		fname = argv[2];
	case 2:
		if (scanInt (argv[1], &lvl) && -1 <= lvl && 9 >= lvl) {
			break;
		}
		if (1 == strlen (argv[1])) {
			lvl = *argv[1];
			break;
		}
	default:
		Tcl_AppendResult (ip,
			"usage: ", argv[0], " <level> [<filename>]", (char*)0);
		return TCL_ERROR;
	}

	openIsisCLog (lvl, fname); /* fname currently ignored */
	return TCL_OK;
}

/*	openIsisCInit(int argc, const char **argv,OpenIsisCLockFunc lock)
	incl. openIsisCOpen ( OpenIsisRec *args )
*/

static int OIT_Init (
	ClientData cd, Tcl_Interp *ip, int argc, const char **argv
) {
	char buf[BUFLEN];
	int  rt;

	rt = openIsisCInit (argc - 1, argv + 1, 0);
	if (0 != rt) {
		PRNBUF ("%d", rt);
		Tcl_SetResult (ip, buf, TCL_VOLATILE);
		return TCL_ERROR;
	}

	return resultSes (ip, OPENISIS_SES0 ());
}

/*	cDOpenv ( const char *dbname, const char **argv, int argc )
*/
static int OIT_Open (
	ClientData cd, Tcl_Interp *ip, int argc, const char **argv
) {
	Tcl_CmdInfo cinfo;
	char  buf[BUFLEN];
	int  *dbid;
	int   rt;

	if (2 > argc) {
		Tcl_AppendResult (ip,
			"usage: ", argv[0], " <dbname> [args...]", (char*)0);
		return TCL_ERROR;
	}

	errno = 0;
	rt = openIsisCDOpenv (argv[1], argv + 2, argc - 2);
	if (0 > rt) {
		if (0 != errno) {
			Tcl_SetResult (ip, strerror (errno), TCL_VOLATILE);
		}
		else {
			PRNBUF ("%d", rt);
			Tcl_SetResult (ip, buf, TCL_VOLATILE);
		}
		return TCL_ERROR;
	}

	PRNBUF ("oidb%d", rt);
	*(dbid = (int*) Tcl_Alloc (sizeof (int))) = rt;
	if (! Tcl_GetCommandInfo (ip, buf, &cinfo)) {
		Tcl_CreateCommand (ip, buf, (Tcl_CmdProc*)OIT_CmdDb,
			(ClientData)dbid, (Tcl_CmdDeleteProc*)OIT_FreeDb);
	}
	Tcl_SetResult (ip, buf, TCL_VOLATILE);
	return TCL_OK;
}


/*	============================ session ================================
*/

static int OIT_CmdSess (
	OpenIsisSession ses, Tcl_Interp *ip, int argc, const char **argv
) {
	if (2 != argc || strncmp ("new", argv[1], strlen (argv[1]))) {
		Tcl_AppendResult (ip,
			"usage: ", argv[0], " new", (char*)0);
		return TCL_ERROR;
	}

	return resultSes (ip, openIsisCSession ((OpenIsisRec*)0));
}

/*	============================ db ================================
*/

static int OIT_CmdDb (
	int *dbid, Tcl_Interp *ip, int argc, const char **argv
) {
	OpenIsisSession ses = (OpenIsisSession)0;
	const char *sname = "";
	const char *rname = "";
	char        buf[BUFLEN];
	unsigned    len;
	int         rt;

	if (2 > argc || 0 == (len = strlen (argv[1]))) {
		goto usage;
	}
	switch (*argv[1]) {

	case 'a':
		if (0 == strncmp ("append", argv[1], len)) {
			Tcl_DString  ds;
			OpenIsisRec *rec;
			long rid;
			int  fidx = 3;

			if (5 > argc || 0 == (argc & 1)) {
				goto usage;
			}
			if (! scanLong (argv[2], &rid)) {
				goto usage;
			}
			if (0 == strncmp ("-session", argv[3], strlen (argv[3]))) {
				if (0 == (ses = scanSes (sname = argv[4]))) {
					goto nosess;
				}
				if ((fidx = 5) == argc) {
					goto usage;
				}
			}
			if (0 == ses) {
				ses = OPENISIS_SES0 ();
			}

			if (0 == (rec = openIsisDRead (ses, *dbid, rid))) {
				rname = argv[2];
				goto norec;
			}
			while (fidx < argc) {
				if (! Append2Rec (ses, &rec, argv[fidx], argv[1 + fidx])) {
					Tcl_AppendResult (ip, argv[0],
						": error appending field ", argv[fidx], (char*)0);
					return TCL_ERROR;
				}
				fidx += 2;
			}

			Tcl_SetResult (ip, Rec2String (rec, &ds), TCL_VOLATILE);
			if (0 != rec) {
				free (rec);
			}
			Tcl_DStringFree (&ds);
			return TCL_OK;

		} /* 'a' */
		goto usage;

	case 'c':
		if (1 != len) {
			if (0 == strncmp ("close", argv[1], len)) {
				if (2 != argc) {
					goto usage;
				}
				Tcl_DeleteCommand (ip, (char*)argv[0]);
				return TCL_OK;
			}
			if (0 == strncmp ("check", argv[1], len)) {
				int flags = 0;
				if (2 != argc) {
					if (3 != argc) {
						goto usage;
					}
					if (! scanInt (argv[2], &flags)) {
						goto usage;
					}
				}
				rt = openIsisCDCheck (*dbid, flags);
				if (0 == rt) {
					return TCL_OK;
				}
				PRNBUF ("%d", rt);
				Tcl_SetResult (ip, buf, TCL_VOLATILE);
				return TCL_ERROR;
			}
		}
		goto usage;

	case 'd':

		/* $db delete <rowid> */
		goto usage;

	case 'i':

		/* $db insert {rec} */
		goto usage;

	case 'm':
		if (0 == strncmp ("maxrowid", argv[1], len)) {
			if (2 != argc) {
				if (4 != argc ||
					strncmp ("-session", argv[2], strlen (argv[2]))) {
					goto usage;
				}
				if (0 == (ses = scanSes (sname = argv[3]))) {
					goto nosess;
				}
			}
			if (0 == ses) {
				ses = OPENISIS_SES0 ();
			}
			rt = openIsisDMaxId (ses, *dbid);
			PRNBUF ("%d", rt);
			if (0 > rt) {
				Tcl_AppendResult (ip,
					"could not determine max rowid: ", buf, (char*)0);
				return TCL_ERROR;
			}
			Tcl_SetResult (ip, buf, TCL_VOLATILE);
			return TCL_OK;
		}
		goto usage;

	case 'q': {
		OIT_Set    *set;
		const char *key;
		long        mode = OPENISIS_QRY_KEYEQ;
		long        skip = 0;
		long        lrt;
		int         j;

		if (strncmp ("query", argv[1], len)) {
			goto usage;
		}
		if (3 > argc) {
			goto usage;
		}
		key = argv[2];
		for (j = 3; argc > j; ++j) {
			if (argv[j][0] != '-') {
				goto usage;
			}
			switch (argv[j][1]) {
			case 'm':
				if (strncmp ("-mode", argv[j], strlen (argv[j]))) {
					goto usage;
				}
				if (argc <= ++j) {
					goto usage;
				}
				if (! scanLong (argv[j], &mode)) {
					goto usage;
				}
				continue;
			case 's': {
				unsigned ll = strlen (argv[j]);
				if (3 > ll) {
					goto usage;
				}
				if (0 == strncmp ("-session", argv[j], ll)) {
					if (argc <= ++j) {
						goto usage;
					}
					if (0 == (ses = scanSes (sname = argv[j]))) {
						goto nosess;
					}
					continue;
				}
				if (0 == strncmp ("-skip", argv[j], ll)) {
					if (argc <= ++j) {
						goto usage;
					}
					if (! scanLong (argv[j], &skip)) {
						goto usage;
					}
					continue;
				}
				} /* fall thru */
			default:
				goto usage;
			}
		} /* for args */

		if (0 == ses) {
			ses = OPENISIS_SES0 ();
		}
		set = (OIT_Set*) Tcl_Alloc (sizeof (OIT_Set));
		set->ois.len = set->idx = 0;
		set->ses = ses;
		set->dbid = *dbid;
		lrt = openIsisDQuery (ses, &(set->ois), *dbid, key, mode, skip);

		if (0 > lrt) {
			Tcl_Free ((ClientData)set);
			PRNBUF ("%ld", lrt);
			Tcl_SetResult (ip, buf, TCL_VOLATILE);
			return TCL_ERROR;
		}

		rt = NewSetCounterFor (set);
		PRNBUF ("oiset%d", rt);
		Tcl_CreateCommand (ip, buf, (Tcl_CmdProc*)OIT_CmdSet,
			(ClientData)set, (Tcl_CmdDeleteProc*)OIT_FreeSet);
		Tcl_SetResult (ip, buf, TCL_VOLATILE);
		return TCL_OK;

		} /* 'q' */
		goto usage;

	case 'r': {
		long rid;

		if (strncmp ("read", argv[1], len)) {
			goto usage;
		}
		switch (argc) {
		case 5:
			if (strncmp ("-session", argv[3], strlen (argv[3]))) {
				goto usage;
			}
			if (0 == (ses = scanSes (sname = argv[4]))) {
				goto nosess;
			}
		case 3:
			if (scanLong (argv[2], &rid) && 0 <= rid) {
				break;
			}
		default:
			goto usage;
		}
		if (0 == ses) {
			ses = OPENISIS_SES0 ();
		}

		return OIT_ReadRec (ip, ses, *dbid, rid);

		} /* 'r' */
		goto usage;

	case 'u':
		if (0 == strncmp ("update", argv[1], len)) {
			Tcl_DString  ds;
			OpenIsisRec *rec;
			long rid;
			int  fidx = 3;

			if (5 > argc || 0 == (argc & 1)) {
				goto usage;
			}
			if (! scanLong (argv[2], &rid)) {
				goto usage;
			}
			if (0 == strncmp ("-session", argv[3], strlen (argv[3]))) {
				if (0 == (ses = scanSes (sname = argv[4]))) {
					goto nosess;
				}
				if ((fidx = 5) == argc) {
					goto usage;
				}
			}
			if (0 == ses) {
				ses = OPENISIS_SES0 ();
			}

			if (0 == (rec = openIsisRDup (ses, (OpenIsisRec*)0, 1024, !0))) {
				Tcl_AppendResult (ip, argv[0],
					": cannot allocate record", (char*)0);
				return TCL_ERROR;
			}
			while (fidx < argc) {
				if (! Append2Rec (ses, &rec, argv[fidx], argv[1 + fidx])) {
					Tcl_AppendResult (ip, argv[0],
						": error appending field ", argv[fidx], (char*)0);
					return TCL_ERROR;
				}
				fidx += 2;
			}

			rec->rowid = rid;
			rt = openIsisDWrite (ses, *dbid, rec);
			if (0 != rt) {
				PRNBUF ("%d", rt); 
				Tcl_SetResult (ip, buf, TCL_VOLATILE);
				return TCL_ERROR;
			}
			Tcl_SetResult (ip, Rec2String (rec, &ds), TCL_VOLATILE);
			if (0 != rec) {
				free (rec);
			}
			Tcl_DStringFree (&ds);
			return TCL_OK;

		} /* 'u' */
		goto usage;

	} /* switch (*argv[1]) */

usage:
	Tcl_AppendResult (ip, "ops on db:\n  ", argv[0],
		" check [<flags>] |",
		" close |",
		" maxrowid [-session <id>] |",
		" query <key> [-session <id>] [-mode <mode>] [-skip <skip>] |",
		" read <rowid> [-session <id>]\n",
/*
		"ops on recs:\n  ", argv[0],
		" append <rowid> [-session <id>] <fields> |",
		" delete <rowid> [-session <id>] |",
		" insert [-session <id>] <fields> |",
		" update [-session <id>] <fields>",
*/
		(char*)0);
	return TCL_ERROR;

nosess:
	Tcl_AppendResult (ip, argv[0], ": no such session: ", sname, (char*)0);
	return TCL_ERROR;

norec:
	Tcl_AppendResult (ip, argv[0], ": no such record: ", rname, (char*)0);
	return TCL_ERROR;
}

static int OIT_FreeDb (
	int *dbid, Tcl_Interp *ip, int argc, const char **argv
) {
	char  buf[BUFLEN];
	int rt;

	rt = openIsisCDClose (*dbid);
	if (0 == rt) {
		Tcl_Free ((ClientData)dbid);
		return TCL_OK;
	}

	PRNBUF2 ("%d: %d", *dbid, rt);
	Tcl_Free ((ClientData)dbid);
	Tcl_AppendResult (ip, "could not close database ", buf, (char*)0);
	return TCL_ERROR;
}

/*	============================ set ================================
*/

static int OIT_CmdSet (
	OIT_Set *set, Tcl_Interp *ip, int argc, const char **argv
) {
	unsigned len;

	switch (argc) {

	case 3: {
		int idx;
		if (0 == (len = strlen (argv[1])) || 'g' != *argv[1]) {
			goto usage;
		}
		if (strncmp ("get", argv[1], len)) {
			goto usage;
		}
		if (! scanInt (argv[2], &idx)) {
			goto usage;
		}
		if (0 > idx || set->ois.len <= idx) {
			char buf[BUFLEN];
			PRNBUF2 ("%d / %ld", idx, set->ois.len);
			Tcl_AppendResult (ip, "index out of bounds: ", buf, (char*)0);
			return TCL_ERROR;
		}
		return OIT_ReadRec (ip, set->ses, set->dbid, set->ois.id[idx]);

		} /* 3 */

	case 2:
		if (0 != (len = strlen (argv[1])) && 'g' != *argv[1]) {
			break;
		}
	default:
		goto usage;

	} /* switch (argc) */

	switch (*argv[1]) {

	case 'c':
		if (0 == strncmp ("close", argv[1], len)) {
			Tcl_DeleteCommand (ip, (char*)argv[0]);
			return TCL_OK;
		}
		break;

	case 'l':
		if (0 == strncmp ("list", argv[1], len)) {
			char   buf[BUFLEN];
			long  *rid;
			int    j;

			for (j = 0, rid = set->ois.id; set->ois.len > j; ++j, ++rid) {
				PRNBUF ("%ld", *rid);
				Tcl_AppendElement (ip, buf);
			}
			return TCL_OK;
		}
		break;

	case 'n':
		if (0 == strncmp ("next", argv[1], len)) {
			if (set->ois.len > set->idx) {
				char   buf[BUFLEN];
				PRNBUF ("%ld", set->ois.id[set->idx++]);
				Tcl_AppendElement (ip, buf);
			}
			return TCL_OK;
		}
		break;

	case 'r':
		if (0 == strncmp ("rewind", argv[1], len)) {
			set->idx = 0;
			return TCL_OK;
		}
		break;

	case 's':
		if (0 == strncmp ("size", argv[1], len)) {
			char  buf[BUFLEN];
			PRNBUF ("%ld", set->ois.len);
			Tcl_SetResult (ip, buf, TCL_VOLATILE);
			return TCL_OK;
		}
		break;

	} /* switch (*argv[1]) */

usage:
	Tcl_AppendResult (ip, "usage: ", argv[0],
		" close | get <index> | list | next | rewind | size",
		(char*)0);
	return TCL_ERROR;
}

static int OIT_FreeSet (
	OIT_Set *set, Tcl_Interp *ip, int argc, const char **argv
) {
	Tcl_Free ((ClientData)set);
	return TCL_OK;
}

/* ************************************************************
	public functions
*/

int
OpenisisTcl_Init ( Tcl_Interp *ip )
{
	Tcl_CreateCommand (ip, "openIsisInit", (Tcl_CmdProc*)OIT_Init,
		(ClientData)0, (Tcl_CmdDeleteProc*)0);
	Tcl_CreateCommand (ip, "openIsisLog", (Tcl_CmdProc*)OIT_Log,
		(ClientData)0, (Tcl_CmdDeleteProc*)0);
	Tcl_CreateCommand (ip, "openIsisOpen", (Tcl_CmdProc*)OIT_Open,
		(ClientData)0, (Tcl_CmdDeleteProc*)0);
	return TCL_OK;
}

int
OpenisisTcl_AppInit ( Tcl_Interp *interp ) {
	if ( TCL_ERROR == OpenisisTcl_Init( interp ) ) return TCL_ERROR;
	return Tcl_AppInit( interp );
}

